"""Input-output utilities."""

import os
import time
from typing import Dict

import h5py
import jax.numpy as jnp
import numpy as np
import pyvista
from omegaconf import DictConfig, OmegaConf


def io_setup(cfg: DictConfig):
    """Setup the output directory and write the arguments to a .txt file."""
    case, solver, dim = cfg.case.name.upper(), cfg.solver.name, cfg.case.dim
    dir = cfg.io.data_path
    if not dir.endswith("/"):
        dir += "/"
    if len(cfg.io.write_type) > 0 and cfg.case.mode == "sim":
        dir += str(dim) + "D_" + case + "_" + solver + "_" + str(cfg.seed)
        dir += "_" + time.strftime("%Y%m%d-%H%M%S")

    os.makedirs(dir, exist_ok=True)
    OmegaConf.save(cfg, os.path.join(dir, "config.yaml"))

    return dir


def write_h5(data_dict: Dict, path: str):
    """Write a dict of numpy or jax arrays to a .h5 file."""
    hf = h5py.File(path, "w")
    for k, v in data_dict.items():
        hf.create_dataset(k, data=np.array(v))
    hf.close()


def write_vtk(data_dict: Dict, path: str):
    """Store a .vtk file for ParaView."""
    data_pv = dict2pyvista(data_dict)
    data_pv.save(path)


def write_state(step: int, state: Dict, dir: str, cfg: DictConfig):
    """Write state to .h5 or .vtk file while simulation is running."""
    step_max = cfg.solver.sequence_length
    write_normal = (
        (cfg.case.mode == "sim") and ((step % cfg.io.write_every) == 0) and (step >= 0)
    )
    write_relax = (cfg.case.mode == "rlx") and (step == (step_max - 1))

    if write_normal or write_relax:
        digits = len(str(step_max))
        step_str = str(step).zfill(digits)

        if cfg.case.mode == "rlx":
            name = [cfg.case.name, str(cfg.case.dim), str(cfg.case.dx), str(cfg.seed)]
            name = "_".join(name)  # e.g. "tgv_3_0.02_42"
        else:
            name = "traj_" + step_str

        if "h5" in cfg.io.write_type:
            path = os.path.join(dir, name + ".h5")
            write_h5(state, path)
        if "vtk" in cfg.io.write_type:
            path = os.path.join(dir, name + ".vtk")
            write_vtk(state, path)


def read_h5(file_name: str, array_type: str = "jax"):
    """Read an .h5 file and return a dict of numpy or jax arrays."""
    hf = h5py.File(file_name, "r")

    data_dict = {}
    for k, v in hf.items():
        if array_type == "jax":
            data_dict[k] = jnp.array(v)
        elif array_type == "numpy":
            data_dict[k] = np.array(v)
        else:
            raise ValueError('array_type must be either "jax" or "numpy"')

    hf.close()

    return data_dict


def write_vtks_from_h5s(dir_path: str, keep_h5: bool = True):
    """Transform a set of .h5 files to .vtk files.

    Args:
        path: path to directory with .h5 files
        keep_h5: Whether to keep or delete the original .h5 files.
    """

    files = os.listdir(dir_path)
    # only regard h5 files
    files = [f for f in files if (".h5" in f)]
    files = sorted(files)

    for i, filename in enumerate(files):
        file_path_h5 = os.path.join(dir_path, filename)
        file_path_vtk = file_path_h5[:-2] + "vtk"
        state = read_h5(file_path_h5)
        write_vtk(state, file_path_vtk)

        if not keep_h5:
            os.remove(file_path_h5)


def render_data_dict(data_dict):
    pl = pyvista.Plotter()
    pl.show_grid()
    pl.add_bounding_box()
    pl.add_axes()
    pl.camera_position = "xy"
    pl.show()


def dict2pyvista(data_dict):
    # PyVista works only with 3D objects, thus we check whether the inputs
    # are 2D and then increase the degrees of freedom of the second dimension.
    # N is the number of points and dim the dimension
    r = np.asarray(data_dict["r"])
    N, dim = r.shape

    # PyVista treats the position information differently than the rest
    if dim == 2:
        r = np.hstack([r, np.zeros((N, 1))])
    data_pv = pyvista.PolyData(r)

    # copy all the other information also to pyvista, using plain numpy arrays
    for k, v in data_dict.items():
        # skip r because we already considered it above
        if k == "r":
            continue

        # working in 3D or scalar features do not require special care
        if dim == 2 and v.ndim == 2:
            v = np.hstack([v, np.zeros((N, 1))])

        data_pv[k] = np.asarray(v)

    return data_pv


if __name__ == "__main__":
    # write_vtks_from_h5s("./3D_TGV_SPH_17_20221224-084051", False)
    pass



"""Weakly compressible SPH solver."""

from typing import Callable, Union

import jax.numpy as jnp
from jax import ops, vmap

from jax_sph.eos import RIEMANNEoS, TaitEoS
from jax_sph.jax_md import space
from jax_sph.kernel import (
    CubicKernel,
    GaussianKernel,
    QuinticKernel,
    SuperGaussianKernel,
    WendlandC2Kernel,
    WendlandC4Kernel,
    WendlandC6Kernel,
)
from jax_sph.utils import Tag, wall_tags

EPS = jnp.finfo(float).eps


def rho_evol_fn(rho, mass, u, grad_w_dist, i_s, j_s, dt, N, **kwargs):
    """Density evolution according to Adami et al. 2013."""
    v_j_s = (mass / rho)[j_s]
    temp = v_j_s * ((u[i_s] - u[j_s]) * grad_w_dist).sum(axis=1)
    drhodt = rho * ops.segment_sum(temp, i_s, N)
    rho = rho + dt * drhodt
    return rho, drhodt


def rho_evol_riemann_fn_wrapper(kernel_fn, eos, c_ref):
    """Density evolution according to Zhang et al. 2017."""

    def rho_evol_riemann_fn(
        e_s,
        rho_i,
        rho_j,
        m_j,
        u_i,
        u_j,
        p_i,
        p_j,
        r_ij,
        d_ij,
        wall_mask_j,
        n_w_j,
        g_ext_i,
        u_tilde_j,
        **kwargs,
    ):
        # Compute unit vector, above eq. (6), Zhang (2017)
        e_ij = e_s

        # Compute kernel gradient
        kernel_grad = kernel_fn.grad_w(d_ij) * (e_ij)

        # Compute average states eq. (6)/(12)/(13), Zhang (2017)
        u_L = jnp.where(
            jnp.isin(wall_mask_j, wall_tags), jnp.dot(u_i, -n_w_j), jnp.dot(u_i, -e_ij)
        )
        p_L = p_i
        rho_L = rho_i

        #  u_w from eq. (15), Yang (2020)
        u_R = jnp.where(
            jnp.isin(wall_mask_j, wall_tags),
            -u_L + 2 * jnp.dot(u_j, n_w_j),
            jnp.dot(u_j, -e_ij),
        )
        p_R = jnp.where(
            jnp.isin(wall_mask_j, wall_tags), p_L + rho_L * jnp.dot(g_ext_i, -r_ij), p_j
        )
        rho_R = jnp.where(jnp.isin(wall_mask_j, wall_tags), eos.rho_fn(p_R), rho_j)

        U_avg = (u_L + u_R) / 2
        v_avg = (u_i + u_j) / 2
        rho_avg = (rho_L + rho_R) / 2

        # Compute Riemann states eq. (7) and below eq. (9), Zhang (2017)
        U_star = U_avg + 0.5 * (p_L - p_R) / (rho_avg * c_ref)
        v_star = U_star * (-e_ij) + (v_avg - U_avg * (-e_ij))

        # Mass conservation with linear Riemann solver eq. (8), Zhang (2017)
        eq_8 = 2 * rho_i * m_j / rho_j * jnp.dot((u_i - v_star), kernel_grad)
        return eq_8

    return rho_evol_riemann_fn


def rho_renorm_fn(rho, mass, i_s, j_s, w_dist, N):
    """Renormalization of density according to Zhang et al. 2017."""
    nominator = ops.segment_sum(mass[j_s] * w_dist, i_s, N)
    rho_denominator = ops.segment_sum((mass / rho)[j_s] * w_dist, i_s, N)
    rho_denominator = jnp.where(rho_denominator > 1, 1, rho_denominator)
    rho = nominator / rho_denominator
    return rho


def rho_summation_fn(mass, i_s, w_dist, N):
    """Density summation."""
    return mass * ops.segment_sum(w_dist, i_s, N)


def wall_phi_vec_wrapper(kernel_fn):
    """Compute the wall phi vector according to Zhang et al. 2017."""

    def wall_phi_vec(rho_j, m_j, dr_ij, dist, tag_j, tag_i):
        # Compute unit vector, above eq. (6), Zhang (2017)
        e_ij_w = dr_ij / (dist + EPS)

        # Compute kernel gradient
        kernel_grad = kernel_fn.grad_w(dist) * (e_ij_w)

        # compute phi eq. (15), Zhang (2017)
        phi = -1.0 * m_j / rho_j * kernel_grad * tag_j * tag_i

        return phi

    return wall_phi_vec


def acceleration_tvf_fn_wrapper(kernel_fn):
    """Transport velocity formulation acceleration according to Adami et al. 2013."""

    def acceleration_tvf_fn(r_ij, d_ij, rho_i, rho_j, m_i, m_j, p_bg_i):
        # compute the common prefactor `_c`
        _weighted_volume = ((m_i / rho_i) ** 2 + (m_j / rho_j) ** 2) / m_i
        _kernel_grad = kernel_fn.grad_w(d_ij)
        _c = _weighted_volume * _kernel_grad / (d_ij + EPS)

        # (Eq. 13) - or at least the acceleration term
        a_eq_13 = _c * 1.0 * p_bg_i * r_ij

        return a_eq_13

    return acceleration_tvf_fn


def tvf_stress_fn(rho: float, u, v):
    """Transport velocity stress tensor. See 'A' under (Eq. 4) in Adami et al. 2013."""
    return jnp.outer(rho * u, v - u)


def acceleration_standard_fn_wrapper(kernel_fn):
    """Standard SPH acceleration according to Adami et al. 2012."""

    def acceleration_standard_fn(
        r_ij,
        d_ij,
        rho_i,
        rho_j,
        u_i,
        u_j,
        v_i,
        v_j,
        m_i,
        m_j,
        eta_i,
        eta_j,
        p_i,
        p_j,
    ):
        # (Eq. 6) - inter-particle-averaged shear viscosity (harmonic mean)
        eta_ij = 2 * eta_i * eta_j / (eta_i + eta_j + EPS)
        # (Eq. 7) - density-weighted pressure (weighted arithmetic mean)
        p_ij = (rho_j * p_i + rho_i * p_j) / (rho_i + rho_j)

        # compute the common prefactor `_c`
        _weighted_volume = ((m_i / rho_i) ** 2 + (m_j / rho_j) ** 2) / m_i
        _kernel_grad = kernel_fn.grad_w(d_ij)
        _c = _weighted_volume * _kernel_grad / (d_ij + EPS)

        # (Eq. 8): \boldsymbol{e}_{ij} is computed as r_ij/d_ij here.
        _A = (tvf_stress_fn(rho_i, u_i, v_i) + tvf_stress_fn(rho_j, u_j, v_j)) / 2
        _u_ij = u_i - u_j
        a_eq_8 = _c * (-p_ij * r_ij + jnp.dot(_A, r_ij) + eta_ij * _u_ij)
        return a_eq_8

    return acceleration_standard_fn


def acceleration_riemann_fn_wrapper(kernel_fn, eos, beta_fn, eta_limiter):
    """Riemann solver acceleration according to Zhang et al. 2017."""

    def acceleration_fn_riemann(
        e_s,
        r_ij,
        d_ij,
        rho_i,
        rho_j,
        m_j,
        m_i,
        u_i,
        u_j,
        p_i,
        p_j,
        eta_i,
        eta_j,
        wall_mask_j,
        mask,
        n_w_j,
        g_ext_i,
        u_tilde_j,
    ):
        # Compute unit vector, above eq. (6), Zhang (2017)
        e_ij = e_s

        # Compute kernel gradient
        kernel_part_diff = kernel_fn.grad_w(d_ij)
        kernel_grad = kernel_part_diff * (e_ij)

        # Compute average states eq. (6)/(12)/(13), Zhang (2017)
        u_L = jnp.where(
            jnp.isin(wall_mask_j, wall_tags), jnp.dot(u_i, -n_w_j), jnp.dot(u_i, -e_ij)
        )
        p_L = p_i
        rho_L = rho_i

        # u_w from eq. (15), Yang (2020)
        u_R = jnp.where(
            jnp.isin(wall_mask_j, wall_tags),
            -u_L + 2 * jnp.dot(u_j, n_w_j),
            jnp.dot(u_j, -e_ij),
        )
        p_R = jnp.where(
            jnp.isin(wall_mask_j, wall_tags), p_L + rho_L * jnp.dot(g_ext_i, -r_ij), p_j
        )
        rho_R = jnp.where(jnp.isin(wall_mask_j, wall_tags), eos.rho_fn(p_R), rho_j)

        P_avg = (p_L + p_R) / 2
        rho_avg = (rho_L + rho_R) / 2

        # Compute inter-particle-averaged shear viscosity (harmonic mean)
        # eq. (6), Adami (2013)
        eta_ij = 2 * eta_i * eta_j / (eta_i + eta_j + EPS)

        # Compute Riemann states eq. (7) and (10), Zhang (2017)
        P_star = P_avg + 0.5 * rho_avg * (u_L - u_R) * beta_fn(u_L, u_R, eta_limiter)

        # pressure term with linear Riemann solver eq. (9), Zhang (2017)
        eq_9 = -2 * m_j * (P_star / (rho_i * rho_j)) * kernel_grad

        # viscosity term eq. (6), Zhang (2019)
        u_d = 2 * u_j - u_tilde_j
        v_ij = jnp.where(
            jnp.isin(wall_mask_j, wall_tags),
            u_i - u_d,
            u_i - u_j,
        )
        eq_6 = 2 * m_j * eta_ij / (rho_i * rho_j) * v_ij / (d_ij + EPS)
        eq_6 *= kernel_part_diff * mask

        # compute the prefactor `_c`
        _weighted_volume = ((m_i / rho_i) ** 2 + (m_j / rho_j) ** 2) / m_i
        _kernel_grad = kernel_fn.grad_w(d_ij)
        _c = _weighted_volume * _kernel_grad / (d_ij + EPS)

        _A = jnp.where(
            jnp.isin(wall_mask_j, wall_tags),
            (tvf_stress_fn(rho_i, u_i, u_i) + tvf_stress_fn(rho_j, u_d, u_d)) / 2,
            (tvf_stress_fn(rho_i, u_i, u_i) + tvf_stress_fn(rho_j, u_j, u_j)) / 2,
        )
        a_eq_8 = _c * jnp.dot(_A, r_ij)

        return eq_9 + eq_6 + a_eq_8

    return acceleration_fn_riemann


def artificial_viscosity_fn_wrapper(dx, artificial_alpha, u_ref=1.0):
    """Artificial viscosity according to Adami et al. 2012."""
    h_ab = dx
    # if only artificial viscosity is used, then the following applies
    # nu = alpha * h * c_ab / 2 / (dim+2)
    #    = 0.1 * 0.02 * 10*1 /2/4= 0.0025
    # TODO: parse reference parameters from case setup
    c_ab = 10.0 * u_ref  # c_ref

    def artificial_viscosity_fn(
        rho, mass, u, tag, i_s, j_s, dr_i_j, dist, grad_w_dist, N
    ):
        rho_ab = (rho[i_s] + rho[j_s]) / 2
        numerator = mass[j_s] * artificial_alpha * h_ab * c_ab
        numerator = numerator * ((u[i_s] - u[j_s]) * dr_i_j).sum(axis=1)
        numerator = numerator[:, None] * grad_w_dist
        denominator = (rho_ab * (dist**2 + 0.01 * h_ab**2))[:, None]

        mask_fluid = tag == Tag.FLUID
        mask_fluid_edges = mask_fluid[j_s] * mask_fluid[i_s]
        res = mask_fluid_edges[:, None] * numerator / denominator
        dudt_artif = ops.segment_sum(res, i_s, N)
        return dudt_artif

    return artificial_viscosity_fn


def gwbc_fn_wrapper(is_free_slip, is_heat_conduction, eos):
    """Enforce wall boundary conditions by treating boundary particles in a special way.

    If solid walls -> apply BC tricks

    Update dummy particles before acceleration computation (if appl.).

    Steps for boundary particles:

    - sum pressure over fluid with sheparding
    - inverse EoS for density
    - sum velocity over fluid with sheparding and * (-1)
    - if free-slip: project velocity onto normal vector
    - subtract that from 2 * u_wall - keeps lid intact

    Based on: "A generalized wall boundary condition for smoothed
    particle hydrodynamics", Adami, Hu, Adams, 2012
    """

    def gwbc_fn(temperature, rho, tag, u, v, p, g_ext, i_s, j_s, w_dist, dr_i_j, nw, N):
        mask_bc = jnp.isin(tag, wall_tags)

        def no_slip_bc_fn(x):
            # for boundary particles, sum over fluid velocities
            x_wall_unnorm = ops.segment_sum(w_j_s_fluid[:, None] * x[j_s], i_s, N)

            # eq. 22 from "A Generalized Wall boundary condition for SPH", 2012
            x_wall = x_wall_unnorm / (w_i_sum_wf[:, None] + EPS)
            # eq. 23 from same paper

            x = jnp.where(mask_bc[:, None], 2 * x - x_wall, x)
            return x

        def free_slip_bc_fn(x, wall_inner_normals):
            # # normal vectors pointing from fluid to wall
            # (1) implement via summing over fluid particles
            # wall_inner = ops.segment_sum(dr_i_j * mask_j_s_fluid[:, None], i_s, N)
            # # (2) implement using color gradient. Requires 2*rc thick wall
            # # wall_inner = - ops.segment_sum(dr_i_j*mask_j_s_wall[:, None], i_s, N)

            # normalization = jnp.sqrt((wall_inner**2).sum(axis=1, keepdims=True))
            # wall_inner_normals = wall_inner / (normalization + EPS)
            # wall_inner_normals = jnp.where(mask_bc[:, None], wall_inner_normals, 0.0)

            # for boundary particles, sum over fluid velocities
            x_wall_unnorm = ops.segment_sum(w_j_s_fluid[:, None] * x[j_s], i_s, N)

            # eq. 22 from "A Generalized Wall boundary condition for SPH", 2012
            x_wall = x_wall_unnorm / (w_i_sum_wf[:, None] + EPS)
            x_wall = wall_inner_normals * (x_wall * wall_inner_normals).sum(
                axis=1, keepdims=True
            )

            # eq. 23 from same paper
            x = jnp.where(mask_bc[:, None], 2 * x - x_wall, x)
            return x

        # require operations with sender fluid and receiver wall/lid
        mask_j_s_fluid = jnp.where(tag[j_s] == Tag.FLUID, 1.0, 0.0)
        w_j_s_fluid = w_dist * mask_j_s_fluid
        # sheparding denominator
        w_i_sum_wf = ops.segment_sum(w_j_s_fluid, i_s, N)

        if is_free_slip:
            # free-slip boundary - ignore viscous interactions with wall
            u = free_slip_bc_fn(u, -nw)
            v = free_slip_bc_fn(v, -nw)
        else:
            # no-slip boundary condition
            u = no_slip_bc_fn(u)
            v = no_slip_bc_fn(v)

        # eq. 27 from "A Generalized Wall boundary condition for SPH", 2012
        # fluid pressure term
        p_wall_unnorm = ops.segment_sum(w_j_s_fluid * p[j_s], i_s, N)

        # external fluid acceleration term
        rho_wf_sum = (rho[j_s] * w_j_s_fluid)[:, None] * dr_i_j
        rho_wf_sum = ops.segment_sum(rho_wf_sum, i_s, N)
        p_wall_ext = (g_ext * rho_wf_sum).sum(axis=1)

        # normalize with sheparding
        p_wall = (p_wall_unnorm + p_wall_ext) / (w_i_sum_wf + EPS)
        p = jnp.where(mask_bc, p_wall, p)

        rho = vmap(eos.rho_fn)(p)

        if is_heat_conduction:
            # wall particles without temperature boundary condition obtain the adjacent
            # fluid temperature
            t_wall_unnorm = ops.segment_sum(w_j_s_fluid * temperature[j_s], i_s, N)
            t_wall = t_wall_unnorm / (w_i_sum_wf + EPS)
            mask = jnp.isin(tag, jnp.array([Tag.SOLID_WALL, Tag.MOVING_WALL]))
            t_wall = jnp.where(mask, t_wall, temperature)

            temperature = t_wall

        return p, rho, u, v, temperature

    return gwbc_fn


def gwbc_fn_riemann_wrapper(is_free_slip, is_heat_conduction):
    """Riemann solver boundary condition for wall particles."""
    if is_free_slip:

        def free_weight(fluid_mask_i, tag_i):
            return fluid_mask_i

        def riemann_velocities(u, w_dist, fluid_mask, i_s, j_s, N):
            return u
    else:

        def free_weight(fluid_mask_i, tag_i):
            return jnp.ones_like(tag_i)

        def riemann_velocities(u, w_dist, fluid_mask, i_s, j_s, N):
            w_dist_fluid = w_dist * fluid_mask[j_s]
            u_wall_nom = ops.segment_sum(w_dist_fluid[:, None] * u[j_s], i_s, N)
            u_wall_denom = ops.segment_sum(w_dist_fluid, i_s, N)
            u_tilde = u_wall_nom / (u_wall_denom[:, None] + EPS)
            return u_tilde

    if is_heat_conduction:

        def heat_bc(mask_j_s_fluid, w_dist, temperature, i_s, j_s, tag, N):
            w_j_s_fluid = w_dist * mask_j_s_fluid
            # sheparding denominator
            w_i_sum_wf = ops.segment_sum(w_j_s_fluid, i_s, N)
            t_wall_unnorm = ops.segment_sum(w_j_s_fluid * temperature[j_s], i_s, N)
            t_wall = t_wall_unnorm / (w_i_sum_wf + EPS)
            mask = jnp.isin(tag, jnp.array([Tag.SOLID_WALL, Tag.MOVING_WALL]))
            t_wall = jnp.where(mask, t_wall, temperature)
            temperature = t_wall
            return temperature
    else:

        def heat_bc(mask_j_s_fluid, w_dist, temperature, i_s, j_s, tag, N):
            return temperature

    return free_weight, riemann_velocities, heat_bc


def limiter_fn_wrapper(eta_limiter, c_ref):
    """if eta_limiter != -1, introduce dissipation limiter eq. (11), Zhang (2017)."""

    if eta_limiter == -1:

        def beta_fn(u_L, u_R, eta_limiter):
            return c_ref
    else:

        def beta_fn(u_L, u_R, eta_limiter):
            temp = eta_limiter * jnp.maximum(u_L - u_R, jnp.zeros_like(u_L))
            beta = jnp.minimum(temp, jnp.full_like(temp, c_ref))
            return beta

    return beta_fn


def temperature_derivative_wrapper(kernel_fn):
    """Temperature derivative according to Cleary 1998."""

    def temperature_derivative(
        e_s, r_ij, d_ij, rho_i, rho_j, m_j, kappa_i, kappa_j, Cp_i, T_i, T_j
    ):
        e_ij = e_s
        _kernel_grad = kernel_fn.grad_w(d_ij)
        _kernel_grad_vector = _kernel_grad * e_ij

        _effective_kappa = (kappa_i * kappa_j) / (kappa_i + kappa_j)
        F_ab = jnp.dot(r_ij, _kernel_grad_vector) / ((d_ij * d_ij) + EPS)  # scalar

        dTdt = (4 * m_j * _effective_kappa * (T_i - T_j) * F_ab) / (
            Cp_i * rho_i * rho_j
        )

        return dTdt

    return temperature_derivative


class WCSPH:
    """Weakly compressible SPH solver with transport velocity formulation."""

    def __init__(
        self,
        displacement_fn: Callable,
        eos: Union[TaitEoS, RIEMANNEoS],
        g_ext_fn: Callable,
        dx: float,
        dim: int,
        dt: float,
        c_ref: float,
        eta_limiter: float = 3,
        solver: str = "SPH",
        kernel: str = "QSK",
        is_bc_trick: bool = False,
        is_rho_evol: bool = False,
        artificial_alpha: float = 0.0,
        is_free_slip: bool = False,
        is_rho_renorm: bool = False,
        is_heat_conduction: bool = False,
    ):
        self.displacement_fn = displacement_fn
        self.solver = solver
        self.g_ext_fn = g_ext_fn
        self.is_bc_trick = is_bc_trick
        self.is_rho_evol = is_rho_evol
        self.is_rho_renorm = is_rho_renorm
        self.dt = dt
        self.eos = eos
        self.artificial_alpha = artificial_alpha
        self.is_heat_conduction = is_heat_conduction

        _beta_fn = limiter_fn_wrapper(eta_limiter, c_ref)
        match kernel:
            case "CSK":
                self._kernel_fn = CubicKernel(h=dx, dim=dim)
            case "QSK":
                self._kernel_fn = QuinticKernel(h=dx, dim=dim)
            case "WC2K":
                self._kernel_fn = WendlandC2Kernel(h=1.3 * dx, dim=dim)
            case "WC4K":
                self._kernel_fn = WendlandC4Kernel(h=1.3 * dx, dim=dim)
            case "WC6K":
                self._kernel_fn = WendlandC6Kernel(h=1.3 * dx, dim=dim)
            case "GK":
                self._kernel_fn = GaussianKernel(h=dx, dim=dim)
            case "SGK":
                self._kernel_fn = SuperGaussianKernel(h=dx, dim=dim)

        self._gwbc_fn = gwbc_fn_wrapper(is_free_slip, is_heat_conduction, eos)
        (
            self._free_weight,
            self._riemann_velocities,
            self._heat_bc,
        ) = gwbc_fn_riemann_wrapper(is_free_slip, is_heat_conduction)
        self._acceleration_tvf_fn = acceleration_tvf_fn_wrapper(self._kernel_fn)
        self._acceleration_riemann_fn = acceleration_riemann_fn_wrapper(
            self._kernel_fn, eos, _beta_fn, eta_limiter
        )
        self._acceleration_fn = acceleration_standard_fn_wrapper(self._kernel_fn)
        self._artificial_viscosity_fn = artificial_viscosity_fn_wrapper(
            dx, artificial_alpha
        )
        self._wall_phi_vec = wall_phi_vec_wrapper(self._kernel_fn)
        self._rho_evol_riemann_fn = rho_evol_riemann_fn_wrapper(
            self._kernel_fn, eos, c_ref
        )
        self._temperature_derivative = temperature_derivative_wrapper(self._kernel_fn)

    def forward_wrapper(self):
        """Wrapper of update step of SPH."""

        def forward(state, neighbors):
            """Update step of SPH solver.

            Args:
                state (dict): Flow fields and particle properties.
                neighbors (_type_): Neighbors object.
            """

            r, tag, mass, eta = state["r"], state["tag"], state["mass"], state["eta"]
            u, v, dudt, dvdt = state["u"], state["v"], state["dudt"], state["dvdt"]
            rho, drhodt, p = state["rho"], state["drhodt"], state["p"]
            nw, kappa, Cp = state["nw"], state["kappa"], state["Cp"]
            temperature, dTdt = state["T"], state["dTdt"]
            N = len(r)

            # precompute displacements `dr` and distances `dist`
            # the second vector is sorted
            i_s, j_s = neighbors.idx
            r_i_s, r_j_s = r[i_s], r[j_s]
            dr_i_j = vmap(self.displacement_fn)(r_i_s, r_j_s)
            dist = space.distance(dr_i_j)
            w_dist = vmap(self._kernel_fn.w)(dist)

            e_s = dr_i_j / (dist[:, None] + EPS)
            # currently only for density evolution and with artificial viscosity
            grad_w_dist_norm = vmap(self._kernel_fn.grad_w)(dist)
            grad_w_dist = grad_w_dist_norm[:, None] * e_s

            # external acceleration field
            g_ext = self.g_ext_fn(r)  # e.g. np.array([[0, -1], [0, -1], ...])

            # masks
            wall_mask = jnp.where(jnp.isin(tag, wall_tags), 1.0, 0.0)
            fluid_mask = jnp.where(tag == Tag.FLUID, 1.0, 0.0)

            ##### Riemann velocity BCs
            if self.is_bc_trick and (self.solver == "RIE"):
                u_tilde = self._riemann_velocities(u, w_dist, fluid_mask, i_s, j_s, N)
            else:
                u_tilde = u

            ##### Density summation or evolution

            # update evolution

            if self.is_rho_evol and (self.solver == "SPH"):
                rho, drhodt = rho_evol_fn(
                    rho, mass, u, grad_w_dist, i_s, j_s, self.dt, N
                )

                if self.is_rho_renorm:
                    rho = rho_renorm_fn(rho, mass, i_s, j_s, w_dist, N)
            elif self.is_rho_evol and (self.solver == "RIE"):
                temp = vmap(self._rho_evol_riemann_fn)(
                    e_s,
                    rho[i_s],
                    rho[j_s],
                    mass[j_s],
                    u[i_s],
                    u[j_s],
                    p[i_s],
                    p[j_s],
                    dr_i_j,
                    dist,
                    wall_mask[j_s],
                    nw[j_s],
                    g_ext[i_s],
                    u_tilde[j_s],
                )
                drhodt = ops.segment_sum(temp, i_s, N) * fluid_mask
                rho = rho + self.dt * drhodt

                if self.is_rho_renorm:
                    rho = rho_renorm_fn(rho, mass, i_s, j_s, w_dist, N)
            else:
                rho_ = rho_summation_fn(mass, i_s, w_dist, N)
                rho = jnp.where(fluid_mask, rho_, rho)

            ##### Compute primitives

            # pressure, and background pressure
            p = vmap(self.eos.p_fn)(rho)
            background_pressure_tvf = vmap(self.eos.p_fn)(jnp.zeros_like(p))

            #####  Apply BC trick

            if self.is_bc_trick and (self.solver == "SPH"):
                p, rho, u, v, temperature = self._gwbc_fn(
                    temperature,
                    rho,
                    tag,
                    u,
                    v,
                    p,
                    g_ext,
                    i_s,
                    j_s,
                    w_dist,
                    dr_i_j,
                    nw,
                    N,
                )
            elif self.is_bc_trick and (self.solver == "RIE"):
                mask = self._free_weight(fluid_mask[i_s], tag[i_s])
                temperature = self._heat_bc(
                    fluid_mask[j_s], w_dist, temperature, i_s, j_s, tag, N
                )
            elif (not self.is_bc_trick) and (self.solver == "RIE"):
                mask = jnp.ones_like(tag[i_s])

            ##### compute heat conduction

            if self.is_heat_conduction:
                # integrate the incomming temperature derivative
                temperature += self.dt * dTdt

                # compute temperature derivative for next step
                out = vmap(self._temperature_derivative)(
                    e_s,
                    dr_i_j,
                    dist,
                    rho[i_s],
                    rho[j_s],
                    mass[j_s],
                    kappa[i_s],
                    kappa[j_s],
                    Cp[i_s],
                    temperature[i_s],
                    temperature[j_s],
                )
                dTdt = ops.segment_sum(out, i_s, N)

            ##### Compute RHS

            if self.solver == "SPH":
                out = vmap(self._acceleration_fn)(
                    dr_i_j,
                    dist,
                    rho[i_s],
                    rho[j_s],
                    u[i_s],
                    u[j_s],
                    v[i_s],
                    v[j_s],
                    mass[i_s],
                    mass[j_s],
                    eta[i_s],
                    eta[j_s],
                    p[i_s],
                    p[j_s],
                )
            elif self.solver == "RIE":
                out = vmap(self._acceleration_riemann_fn)(
                    e_s,
                    dr_i_j,
                    dist,
                    rho[i_s],
                    rho[j_s],
                    mass[j_s],
                    mass[i_s],
                    u[i_s],
                    u[j_s],
                    p[i_s],
                    p[j_s],
                    eta[i_s],
                    eta[j_s],
                    wall_mask[j_s],
                    mask,
                    nw[j_s],
                    g_ext[i_s],
                    u_tilde[j_s],
                )
            dudt = ops.segment_sum(out, i_s, N)

            out_tv = vmap(self._acceleration_tvf_fn)(
                dr_i_j,
                dist,
                rho[i_s],
                rho[j_s],
                mass[i_s],
                mass[j_s],
                background_pressure_tvf[i_s],
            )
            dvdt = ops.segment_sum(out_tv, i_s, N)

            ##### Additional things

            if self.artificial_alpha != 0.0:
                dudt += self._artificial_viscosity_fn(
                    rho, mass, u, tag, i_s, j_s, dr_i_j, dist, grad_w_dist, N
                )

            state = {
                "r": r,
                "tag": tag,
                "u": u,
                "v": v,
                "drhodt": drhodt,
                "dudt": dudt + g_ext,
                "dvdt": dvdt,
                "rho": rho,
                "p": p,
                "mass": mass,
                "eta": eta,
                "dTdt": dTdt,
                "T": temperature,
                "kappa": kappa,
                "Cp": Cp,
                "nw": nw,
            }

            return state

        return forward



# Source: https://github.com/jax-md/jax-md
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines utility functions."""

from functools import partial
from typing import Any

import jax.numpy as jnp
import numpy as onp
from jax import jit

Array = Any
PyTree = Any

i16 = jnp.int16
i32 = jnp.int32
i64 = jnp.int64

f32 = jnp.float32
f64 = jnp.float64


@partial(jit, static_argnums=(1,))
def safe_mask(mask, fn, operand, placeholder=0):
    masked = jnp.where(mask, operand, 0)
    return jnp.where(mask, fn(masked), placeholder)


def is_array(x: Any) -> bool:
    return isinstance(x, (jnp.ndarray, onp.ndarray))



"""Neighbors search backends."""

from functools import partial
from typing import Optional

import jax
import jax.lax as lax
import jax.numpy as jnp
import numpy as np
import numpy as onp
from jax import jit

from jax_sph.jax_md import space
from jax_sph.jax_md.partition import (
    MaskFn,
    NeighborFn,
    NeighborList,
    NeighborListFns,
    NeighborListFormat,
    PartitionError,
    PartitionErrorCode,
    _displacement_or_metric_to_metric_sq,
    _neighboring_cells,
    cell_list,
    is_format_valid,
    is_sparse,
    shift_array,
)
from jax_sph.jax_md.partition import neighbor_list as vmap_neighbor_list

PEC = PartitionErrorCode


def get_particle_cells(idx, cl_capacity, N):
    """
    Given a cell list idx of shape (nx, ny, nz, cell_capacity), we first
    enumerate each cell and then return a list of shape (N,) containing the
    number of the cell each particle belongs to.
    """
    # containes particle indices in each cell (num_cells, cell_capacity)
    idx = idx.reshape(-1, cl_capacity)

    # (num_cells, cell_capacity) of
    # [[0,0,...0],[1,1,...1],...,[num_cells-1,num_cells-1,...num_cells-1]
    list_cells = jnp.broadcast_to(jnp.arange(idx.shape[0])[:, None], idx.shape)

    idx = jnp.reshape(idx, (-1,))  # flatten
    list_cells = jnp.reshape(list_cells, (-1,))  # flatten

    ordering = jnp.argsort(idx)  # each particle is only once in the cell list
    particle_cells = list_cells[ordering][:N]
    return particle_cells


def _scan_neighbor_list(
    displacement_or_metric: space.DisplacementOrMetricFn,
    box: space.Box,
    r_cutoff: float,
    dr_threshold: float = 0.0,
    capacity_multiplier: float = 1.25,
    disable_cell_list: bool = False,
    mask_self: bool = True,
    custom_mask_function: Optional[MaskFn] = None,
    fractional_coordinates: bool = False,
    format: NeighborListFormat = NeighborListFormat.Sparse,
    num_partitions: int = 8,
    **static_kwargs,
) -> NeighborFn:
    """Modified JAX-MD neighbor list function that uses `lax.scan` to compute the
    distance between particles to save memory.

    Original: https://github.com/jax-md/jax-md/blob/main/jax_md/partition.py

    Returns a function that builds a list neighbors for collections of points.

    Neighbor lists must balance the need to be jit compatible with the fact that
    under a jit the maximum number of neighbors cannot change (owing to static
    shape requirements). To deal with this, our `neighbor_list` returns a
    `NeighborListFns` object that contains two functions: 1)
    `neighbor_fn.allocate` create a new neighbor list and 2) `neighbor_fn.update`
    updates an existing neighbor list. Neighbor lists themselves additionally
    have a convenience `update` member function.

    Note that allocation of a new neighbor list cannot be jit compiled since it
    uses the positions to infer the maximum number of neighbors (along with
    additional space specified by the `capacity_multiplier`). Updating the
    neighbor list can be jit compiled; if the neighbor list capacity is not
    sufficient to store all the neighbors, the `did_buffer_overflow` bit
    will be set to `True` and a new neighbor list will need to be reallocated.

    Here is a typical example of a simulation loop with neighbor lists:

    .. code-block:: python

        init_fn, apply_fn = simulate.nve(energy_fn, shift, 1e-3)
        exact_init_fn, exact_apply_fn = simulate.nve(exact_energy_fn, shift, 1e-3)

        nbrs = neighbor_fn.allocate(R)
        state = init_fn(random.PRNGKey(0), R, neighbor_idx=nbrs.idx)

        def body_fn(i, state):
        state, nbrs = state
        nbrs = nbrs.update(state.position)
        state = apply_fn(state, neighbor_idx=nbrs.idx)
        return state, nbrs

        step = 0
        for _ in range(20):
        new_state, nbrs = lax.fori_loop(0, 100, body_fn, (state, nbrs))
        if nbrs.did_buffer_overflow:
            nbrs = neighbor_fn.allocate(state.position)
        else:
            state = new_state
            step += 1

    Args:
        displacement: A function `d(R_a, R_b)` that computes the displacement
        between pairs of points.
        box: Either a float specifying the size of the box or an array of
        shape `[spatial_dim]` specifying the box size in each spatial dimension.
        r_cutoff: A scalar specifying the neighborhood radius.
        dr_threshold: A scalar specifying the maximum distance particles can move
        before rebuilding the neighbor list.
        capacity_multiplier: A floating point scalar specifying the fractional
        increase in maximum neighborhood occupancy we allocate compared with the
        maximum in the example positions.
        disable_cell_list: An optional boolean. If set to `True` then the neighbor
        list is constructed using only distances. This can be useful for
        debugging but should generally be left as `False`.
        mask_self: An optional boolean. Determines whether points can consider
        themselves to be their own neighbors.
        custom_mask_function: An optional function. Takes the neighbor array
        and masks selected elements. Note: The input array to the function is
        `(n_particles, m)` where the index of particle 1 is in index in the first
        dimension of the array, the index of particle 2 is given by the value in
        the array
        fractional_coordinates: An optional boolean. Specifies whether positions
        will be supplied in fractional coordinates in the unit cube, :math:`[0, 1]^d`.
        If this is set to True then the `box_size` will be set to `1.0` and the
        cell size used in the cell list will be set to `cutoff / box_size`.
        format: The format of the neighbor list; see the :meth:`NeighborListFormat` enum
        for details about the different choices for formats. Defaults to `Dense`.
        **static_kwargs: kwargs that get threaded through the calculation of
        example positions.
    Returns:
        A NeighborListFns object that contains a method to allocate a new neighbor
        list and a method to update an existing neighbor list.
    """
    assert disable_cell_list is False, "Works only with a cell list"
    assert not fractional_coordinates, "Works only with real coordinates"
    assert format == NeighborListFormat.Sparse, "Works only with sparse neighbor list"
    assert custom_mask_function is None, "Custom masking not implemented"

    is_format_valid(format)
    box = lax.stop_gradient(box)
    r_cutoff = lax.stop_gradient(r_cutoff)
    dr_threshold = lax.stop_gradient(dr_threshold)

    box = jnp.float32(box)

    cutoff = r_cutoff + dr_threshold
    cutoff_sq = cutoff**2
    threshold_sq = (dr_threshold / jnp.float32(2)) ** 2
    metric_sq = _displacement_or_metric_to_metric_sq(displacement_or_metric)

    cell_size = cutoff
    assert jnp.all(cell_size < box / 3.0), "Don't use scan with very few cells"

    def neighbor_list_fn(
        position: jnp.ndarray,
        neighbors: Optional[NeighborList] = None,
        extra_capacity: int = 0,
        **kwargs,
    ) -> NeighborList:
        def neighbor_fn(position_and_error, max_occupancy=None):
            position, err = position_and_error
            N, dim = position.shape
            cl_fn = None
            cl = None
            cell_size = None

            if neighbors is None:  # cl.shape = (nx, ny, nz, cell_capacity, dim)
                cell_size = cutoff
                cl_fn = cell_list(box, cell_size, capacity_multiplier)
                cl = cl_fn.allocate(position, extra_capacity=extra_capacity)
            else:
                cell_size = neighbors.cell_size
                cl_fn = neighbors.cell_list_fn
                if cl_fn is not None:
                    cl = cl_fn.update(position, neighbors.cell_list_capacity)

            err = err.update(PEC.CELL_LIST_OVERFLOW, cl.did_buffer_overflow)
            cl_capacity = cl.cell_capacity

            idx = cl.id_buffer

            cell_idx = [idx]  # shape: (nx, ny, nz, cell_capacity, 1)

            for dindex in _neighboring_cells(dim):
                if onp.all(dindex == 0):
                    continue
                cell_idx += [shift_array(idx, dindex)]

            cell_idx = jnp.concatenate(cell_idx, axis=-2)
            cell_idx = jnp.reshape(cell_idx, (-1, cell_idx.shape[-2]))
            num_cells, considered_neighbors = cell_idx.shape

            particle_cells = get_particle_cells(idx, cl_capacity, N)

            d = partial(metric_sq, **kwargs)
            d = space.map_bond(d)

            # number of particles per partition N_sub
            # np.ceil used to pad last partition with < num_partitions entries
            N_sub = int(np.ceil(N / num_partitions))
            num_pad = N_sub * num_partitions - N
            particle_cells = jnp.pad(
                particle_cells,
                (
                    0,
                    num_pad,
                ),
                constant_values=-1,
            )

            if dim == 2:
                # the area of a circle with r=1/3 is 0.34907
                volumetric_factor = 0.34907
            elif dim == 3:
                # the volume of a sphere with r=1/3 is 0.15514
                volumetric_factor = 0.15514

            num_edges_sub = int(
                N_sub * considered_neighbors * volumetric_factor * capacity_multiplier
            )

            def scan_body(carry, input):
                """Compute neighbors over a subset of particles

                The largest object here is of size (N_sub*considered_neighbors), where
                considered_neighbors in 3D is 27 * cell_capacity.
                """

                occupancy = carry
                slice_from = input

                _entries = lax.dynamic_slice(particle_cells, (slice_from,), (N_sub,))
                _idx = cell_idx[_entries]

                if mask_self:
                    particle_idx = slice_from + jnp.arange(N_sub)
                    _idx = jnp.where(_idx == particle_idx[:, None], N, _idx)

                if num_pad > 0:
                    _idx = jnp.where(_entries[:, None] != -1, _idx, N)

                sender_idx = (
                    jnp.broadcast_to(
                        jnp.arange(N_sub, dtype="int32")[:, None], _idx.shape
                    )
                    + slice_from
                )
                if num_pad > 0:
                    sender_idx = jnp.clip(sender_idx, a_max=N)

                sender_idx = jnp.reshape(sender_idx, (-1,))
                receiver_idx = jnp.reshape(_idx, (-1,))
                dR = d(position[sender_idx], position[receiver_idx])

                mask = (dR < cutoff_sq) & (receiver_idx < N)
                out_idx = N * jnp.ones(receiver_idx.shape, jnp.int32)

                cumsum = jnp.cumsum(mask)
                index = jnp.where(mask, cumsum - 1, considered_neighbors * N - 1)
                receiver_idx = out_idx.at[index].set(receiver_idx)
                sender_idx = out_idx.at[index].set(sender_idx)
                occupancy += cumsum[-1]

                carry = occupancy
                y = jnp.stack(
                    (receiver_idx[:num_edges_sub], sender_idx[:num_edges_sub])
                )
                overflow = cumsum[-1] > num_edges_sub
                return carry, (y, overflow)

            carry = jnp.array(0)
            xs = jnp.array([i * N_sub for i in range(num_partitions)])
            occupancy, (idx, overflows) = lax.scan(
                scan_body, carry, xs, length=num_partitions
            )
            err = err.update(PEC.CELL_LIST_OVERFLOW, overflows.sum())
            idx = idx.transpose(1, 2, 0).reshape(2, -1)

            # sort to enable pruning later
            ordering = jnp.argsort(idx[1])
            idx = idx[:, ordering]

            if max_occupancy is None:
                _extra_capacity = N * extra_capacity
                max_occupancy = int(occupancy * capacity_multiplier + _extra_capacity)
                if max_occupancy > idx.shape[-1]:
                    max_occupancy = idx.shape[-1]
                if not is_sparse(format):
                    capacity_limit = N - 1 if mask_self else N
                elif format is NeighborListFormat.Sparse:
                    capacity_limit = N * (N - 1) if mask_self else N**2
                else:
                    capacity_limit = N * (N - 1) // 2
                if max_occupancy > capacity_limit:
                    max_occupancy = capacity_limit
            idx = idx[:, :max_occupancy]
            update_fn = neighbor_list_fn if neighbors is None else neighbors.update_fn
            return NeighborList(
                idx,
                position,
                err.update(PEC.NEIGHBOR_LIST_OVERFLOW, occupancy > max_occupancy),
                cl_capacity,
                max_occupancy,
                format,
                cell_size,
                cl_fn,
                update_fn,
            )  # pytype: disable=wrong-arg-count

        nbrs = neighbors
        if nbrs is None:
            return neighbor_fn((position, PartitionError(jnp.zeros((), jnp.uint8))))

        neighbor_fn = partial(neighbor_fn, max_occupancy=nbrs.max_occupancy)

        d = partial(metric_sq, **kwargs)
        d = jax.vmap(d)

        return lax.cond(
            jnp.any(d(position, nbrs.reference_position) > threshold_sq),
            (position, nbrs.error),
            neighbor_fn,
            nbrs,
            lambda x: x,
        )

    def allocate_fn(
        position: jnp.ndarray, extra_capacity: int = 0, **kwargs
    ) -> NeighborList:
        return neighbor_list_fn(position, extra_capacity=extra_capacity, **kwargs)

    @jit
    def update_fn(
        position: jnp.ndarray, neighbors: NeighborList, **kwargs
    ) -> NeighborList:
        return neighbor_list_fn(position, neighbors, **kwargs)

    return NeighborListFns(allocate_fn, update_fn)  # pytype: disable=wrong-arg-count


def _matscipy_neighbor_list(
    displacement_or_metric: space.DisplacementOrMetricFn,
    box_size: space.Box,
    r_cutoff: float,
    dr_threshold: float = 0.0,
    capacity_multiplier: float = 1.25,
    disable_cell_list: bool = False,
    mask_self: bool = True,
    custom_mask_function: Optional[MaskFn] = None,
    fractional_coordinates: bool = False,
    format: NeighborListFormat = NeighborListFormat.Dense,
    **static_kwargs,
) -> NeighborFn:
    pbc = static_kwargs["pbc"]
    num_particles_max = static_kwargs["num_particles_max"]

    from matscipy.neighbours import neighbour_list as matscipy_nl

    assert box_size.ndim == 1 and (len(box_size) in [2, 3])
    if box_size.shape == (2,):
        box_size = np.pad(box_size, (0, 1), mode="constant", constant_values=1.0)
    if box_size.shape != (3, 3):
        box_size = np.diag(box_size)

    if len(pbc) == 2:
        pbc = np.pad(pbc, (0, 1), mode="constant", constant_values=False)
    else:
        pbc = np.asarray(pbc, dtype=bool)

    dtype_idx = jnp.arange(0).dtype  # just to get the correct dtype

    def matscipy_wrapper(position, idx_shape, num_particles):
        position = position[:num_particles]

        if position.shape[1] == 2:
            position = np.pad(
                position, ((0, 0), (0, 1)), mode="constant", constant_values=0.5
            )

        edge_list = matscipy_nl(
            "ij", cutoff=r_cutoff, positions=position, cell=box_size, pbc=pbc
        )
        edge_list = np.asarray(edge_list, dtype=dtype_idx)
        if not mask_self:
            # add self connection, which matscipy does not do
            self_connect = np.arange(num_particles, dtype=dtype_idx)
            self_connect = np.array([self_connect, self_connect])
            edge_list = np.concatenate((self_connect, edge_list), axis=-1)

        if edge_list.shape[1] > idx_shape[1]:  # overflow true case
            idx_new = np.asarray(edge_list[:, : idx_shape[1]])
            buffer_overflow = np.array(True)
        else:
            idx_new = np.ones(idx_shape, dtype=dtype_idx) * num_particles_max
            idx_new[:, : edge_list.shape[1]] = edge_list
            buffer_overflow = np.array(False)

        return idx_new, buffer_overflow

    @jax.jit
    def update_fn(
        position: jnp.ndarray, neighbors: NeighborList, **kwargs
    ) -> NeighborList:
        num_particles = kwargs["num_particles"]

        shape_edgelist = jax.ShapeDtypeStruct(
            neighbors.idx.shape, dtype=neighbors.idx.dtype
        )
        shape_overflow = jax.ShapeDtypeStruct((), dtype=bool)
        shape_out = (shape_edgelist, shape_overflow)
        idx, buffer_overflow = jax.pure_callback(
            matscipy_wrapper, shape_out, position, neighbors.idx.shape, num_particles
        )

        return NeighborList(
            idx,
            position,
            neighbors.error.update(PEC.NEIGHBOR_LIST_OVERFLOW, buffer_overflow),
            None,
            None,
            None,
            None,
            None,
            update_fn,
        )

    def allocate_fn(
        position: jnp.ndarray, extra_capacity: int = 0, **kwargs
    ) -> NeighborList:
        num_particles = kwargs["num_particles"]
        position = position[:num_particles]

        if position.shape[1] == 2:
            position = np.pad(
                position, ((0, 0), (0, 1)), mode="constant", constant_values=0.5
            )

        edge_list = matscipy_nl(
            "ij", cutoff=r_cutoff, positions=position, cell=box_size, pbc=pbc
        )
        edge_list = jnp.asarray(edge_list, dtype=dtype_idx)
        if not mask_self:
            # add self connection, which matscipy does not do
            self_connect = jnp.arange(num_particles, dtype=dtype_idx)
            self_connect = jnp.array([self_connect, self_connect])
            edge_list = jnp.concatenate((self_connect, edge_list), axis=-1)

        # in case this is a (2,M) pair list, we pad with N and capacity_multiplier
        factor = capacity_multiplier * num_particles_max / num_particles
        res = num_particles * jnp.ones(
            (2, round(edge_list.shape[1] * factor + extra_capacity)),
            dtype_idx,
        )
        res = res.at[:, : edge_list.shape[1]].set(edge_list)
        return NeighborList(
            res,
            position,
            PartitionError(jnp.zeros((), jnp.uint8)),
            None,
            None,
            None,
            None,
            None,
            update_fn,
        )

    return NeighborListFns(allocate_fn, update_fn)


BACKENDS = {
    "jaxmd_vmap": vmap_neighbor_list,
    "jaxmd_scan": _scan_neighbor_list,
    "matscipy": _matscipy_neighbor_list,
}


def neighbor_list(
    displacement_or_metric: space.DisplacementOrMetricFn,
    box_size: space.Box,
    r_cutoff: float,
    backend: str = "jaxmd_vmap",
    dr_threshold: float = 0.0,
    capacity_multiplier: float = 1.25,
    disable_cell_list: bool = False,
    mask_self: bool = True,
    custom_mask_function: Optional[MaskFn] = None,
    fractional_coordinates: bool = False,
    format: NeighborListFormat = NeighborListFormat.Sparse,
    num_particles_max: int = None,
    num_partitions: int = 1,
    pbc: jnp.ndarray = None,
) -> NeighborFn:
    """Neighbor lists wrapper. Its arguments are mainly based on the jax-md ones.

    Args:
        displacement: A function `d(R_a, R_b)` that computes the displacement
            between pairs of points.
        box_size: Either a float specifying the size of the box or an array of
            shape `[spatial_dim]` specifying the box size in each spatial dimension.
        r_cutoff: A scalar specifying the neighborhood radius.
        dr_threshold: A scalar specifying the maximum distance particles can move
            before rebuilding the neighbor list.
        backend: The backend to use. Can be one of: 1) ``jaxmd_vmap`` - the default
            jax-md neighbor list which vectorizes the computations. 2) ``jaxmd_scan`` -
            a modified jax-md neighbor list which serializes the search into
            ``num_partitions`` chunks to improve the memory efficiency. 3) ``matscipy``
            - a jit-able implementation with the matscipy neighbor list backend, which
            runs on CPU and takes variable number of particles smaller or equal to
            ``num_particles``.
        capacity_multiplier: A floating point scalar specifying the fractional
            increase in maximum neighborhood occupancy we allocate compared with the
            maximum in the example positions.
        disable_cell_list: An optional boolean. If set to `True` then the neighbor
            list is constructed using only distances. This can be useful for
            debugging but should generally be left as `False`.
        mask_self: An optional boolean. Determines whether points can consider
            themselves to be their own neighbors.
        custom_mask_function: An optional function. Takes the neighbor array
            and masks selected elements. Note: The input array to the function is
            `(n_particles, m)` where the index of particle 1 is in index in the first
            dimension of the array, the index of particle 2 is given by the value in
            the array
        fractional_coordinates: An optional boolean. Specifies whether positions will
            be supplied in fractional coordinates in the unit cube, :math:`[0, 1]^d`.
            If this is set to True then the `box_size` will be set to `1.0` and the
            cell size used in the cell list will be set to `cutoff / box_size`.
        format: The format of the neighbor list; see the :meth:`NeighborListFormat` enum
            for details about the different choices for formats. Defaults to `Dense`.
        num_particles_max: only used with the ``matscipy`` backend. Based
            on the largest particles system in a dataset.
        num_partitions: only used with the ``jaxmd_scan`` backend
        pbc: only used with the ``matscipy`` backend. Defines the boundary conditions
            for each dimension individually. Can have shape (2,) or (3,).
        **static_kwargs: kwargs that get threaded through the calculation of
            example positions.
    Returns:
        A NeighborListFns object that contains a method to allocate a new neighbor
        list and a method to update an existing neighbor list.
    """
    assert backend in BACKENDS, f"Unknown backend {backend}"

    return BACKENDS[backend](
        displacement_or_metric,
        box_size,
        r_cutoff,
        dr_threshold,
        capacity_multiplier,
        disable_cell_list,
        mask_self,
        custom_mask_function,
        fractional_coordinates,
        format,
        num_particles_max=num_particles_max,
        num_partitions=num_partitions,
        pbc=pbc,
    )


if __name__ == "__main__":
    # edge_list = matscipy_nl(
    #     "ij",
    #     cutoff=0.45,
    #     positions=np.array([[0.1, 0.1, 0.1], [0.5, 0.1, 0.1], [0.5, 0.5, 0.1]]),
    #     cell=np.array([[1.0, 0.0, 0.0],[0.0, 1.0, 0.0], [0.0, 0.0, 1.0]]),
    #     pbc=np.array([True, True, True])
    # )

    # box_size = np.array([1.0, 1.0])
    # r = np.array([[0.1, 0.1], [0.5, 0.1]])
    # displacement_fn, shift_fn = space.periodic(side=box_size)
    # r_cutoff = 0.45

    # neighbor_fn = neighbor_list(
    #     displacement_fn,
    #     box_size,
    #     r_cutoff=r_cutoff,
    #     backend="jaxmd_vmap",
    #     dr_threshold=r_cutoff * 0.25,
    #     capacity_multiplier=1.25,
    #     mask_self=False,
    #     format=Sparse,
    # )
    # neighbors = neighbor_fn.allocate(r)
    # neighbors = neighbors.update(r)
    # neighbors = neighbors.update(r)

    # a = np.array([1,2,3])
    # f = lambda x: x**2

    # def distance_fn(x, y):
    #     return lax.scan(lambda _, x: (None, d(*x)), None, (x, y))[1]

    # def scan(f, init, xs, length=None):
    #     if xs is None:
    #         xs = [None] * length
    #     carry = init
    #     ys = []
    #     for x in xs:
    #         carry, y = f(carry, x)
    #         ys.append(y)
    #     return carry, np.stack(ys)

    pass



"""Integrator schemes."""

from typing import Callable, Dict

from jax_sph.utils import Tag


def si_euler(
    tvf: float, model: Callable, shift_fn: Callable, bc_fn: Callable, nw_fn: Callable
):
    """Semi-implicit Euler integrator including Transport Velocity.

    The integrator advances the state of the system following the steps:

        1. Update u and v with dudt and dvdt
        2. Update position r with velocity v
        3. Update neighbor list
        4. Compute accelerations - SPH model call
        5. Impose boundary conditions as defined by the case.
    """

    def advance(dt: float, state: Dict, neighbors):
        """Call to integrator."""

        # 1. Twice 1/2dt integration of u and v
        state["u"] += 1.0 * dt * state["dudt"]
        state["v"] = state["u"] + tvf * 0.5 * dt * state["dvdt"]

        # 2. Integrate position with velocity v
        state["r"] = shift_fn(state["r"], 1.0 * dt * state["v"])

        # recompute wall normals if needed
        if nw_fn is not None:
            state["nw"] = nw_fn(state["r"])

        # 3. Update neighbor list

        # The displacment and shift function from JAX MD are used for computing the
        # relative particle distance and moving particles across the priodic domain
        #
        # dr = displacement_fn(r1, r2) = r1 - r2
        # respecting PBC, i.e. returns |dr_i| < box_size / 2
        #
        # r = shift_fn(r, dr) = r + dr
        # respecting PBC, i.e. new r in [0, box_size]

        num_particles = (state["tag"] != Tag.PAD_VALUE).sum()
        neighbors = neighbors.update(state["r"], num_particles=num_particles)

        # 4. Compute accelerations
        state = model(state, neighbors)

        # 5. Impose boundary conditions on dummy particles (if applicable)
        state = bc_fn(state)

        return state, neighbors

    return advance



"""Core simulation loop."""

import os
import time

import numpy as np
from jax import jit
from omegaconf import DictConfig, OmegaConf

from jax_sph import partition
from jax_sph.case_setup import load_case, set_relaxation
from jax_sph.integrator import si_euler
from jax_sph.io_state import io_setup, write_state
from jax_sph.jax_md.partition import Sparse
from jax_sph.solver import WCSPH
from jax_sph.utils import Logger, Tag


def simulate(cfg: DictConfig):
    """Core simulation loop.

    This function is the main entry point for the simulation and does the following:

    1. Initialize the 1) case, 2) solver, 3) neighbor list, 4) integrator.
    2. Run the simulation loop and optionally write states.
    """

    # Case setup
    Case = load_case(os.path.dirname(cfg.config), cfg.case.source)
    # if in relaxation mode, wrap the case with the relaxation case
    if cfg.case.mode == "rlx":
        case = set_relaxation(Case, cfg)
    elif cfg.case.mode == "sim":
        case = Case(cfg)
    (
        cfg,
        box_size,
        state,
        g_ext_fn,
        bc_fn,
        nw_fn,
        eos_fn,
        key,
        displacement_fn,
        shift_fn,
    ) = case.initialize()

    # Solver setup
    solver = WCSPH(
        displacement_fn,
        eos_fn,
        g_ext_fn,
        cfg.case.dx,
        cfg.case.dim,
        cfg.solver.dt,
        cfg.case.c_ref,
        cfg.solver.eta_limiter,
        cfg.solver.name,
        cfg.kernel.name,
        cfg.solver.is_bc_trick,
        cfg.solver.density_evolution,
        cfg.solver.artificial_alpha,
        cfg.solver.free_slip,
        cfg.solver.density_renormalize,
        cfg.solver.heat_conduction,
    )
    forward = solver.forward_wrapper()

    # Initialize a neighbors list for looping through the local neighborhood
    # cell_size = r_cutoff + dr_threshold
    # capacity_multiplier is used for preallocating the (2, NN) neighbors.idx
    neighbor_fn = partition.neighbor_list(
        displacement_fn,
        box_size,
        r_cutoff=solver._kernel_fn.cutoff,
        backend=cfg.nl.backend,
        capacity_multiplier=1.25,
        mask_self=False,
        format=Sparse,
        num_particles_max=state["r"].shape[0],
        num_partitions=cfg.nl.num_partitions,
        pbc=np.array(cfg.case.pbc),
    )
    num_particles = (state["tag"] != Tag.PAD_VALUE).sum()
    neighbors = neighbor_fn.allocate(state["r"], num_particles=num_particles)

    # Instantiate advance function for our use case
    advance = si_euler(cfg.solver.tvf, forward, shift_fn, bc_fn, nw_fn)

    advance = advance if cfg.no_jit else jit(advance)

    print("#" * 79, "\nStarting a JAX-SPH run with the following configs:")
    print(OmegaConf.to_yaml(cfg))
    print("#" * 79)

    # create data directory and dump config.yaml
    dir = io_setup(cfg)
    # set up progress bar logging
    logger = Logger(
        dt=cfg.solver.dt,
        dx=cfg.case.dx,
        print_props=cfg.io.print_props,
        sequence_length=cfg.solver.sequence_length,
    )

    # compile kernel and initialize accelerations
    _state, _neighbors = advance(0.0, state, neighbors)
    _state["v"].block_until_ready()

    start = time.time()
    for step in range(cfg.solver.sequence_length + 2):
        write_state(step - 1, state, dir, cfg)

        state_, neighbors_ = advance(cfg.solver.dt, state, neighbors)

        # Check whether the edge list is too small and if so, create longer one
        if neighbors_.did_buffer_overflow:
            edges_ = neighbors.idx.shape
            print(f"Reallocate neighbors list {edges_} at step {step}")
            neighbors = neighbor_fn.allocate(state["r"], num_particles=num_particles)
            print(f"To list {neighbors.idx.shape}")

            # To run the loop N times even if sometimes did_buffer_overflow > 0
            # we directly rerun the advance step here
            state, neighbors = advance(cfg.solver.dt, state, neighbors)
        else:
            state, neighbors = state_, neighbors_

        # update the progress bar
        if step % cfg.io.write_every == 0:
            logger.print_stats(state, step)

    print(f"time: {time.time() - start:.2f} s")



"""Simulation setup."""

import importlib
import os
import warnings
from abc import ABC, abstractmethod

import jax
import jax.numpy as jnp
import numpy as np
from jax import vmap

from jax_sph.eos import RIEMANNEoS, TaitEoS
from jax_sph.io_state import read_h5
from jax_sph.jax_md import space
from jax_sph.utils import (
    Tag,
    compute_nws_jax_wrapper,
    compute_nws_scipy,
    get_noise_masked,
    pos_init_cartesian_2d,
    pos_init_cartesian_3d,
    wall_tags,
)

EPS = jnp.finfo(float).eps


class SimulationSetup(ABC):
    """Parent class for all simulation setups."""

    def __init__(self, cfg):
        # If a value is changed in one of these, it also changes in the others
        self.cfg = cfg
        self.case = cfg.case
        self.special = cfg.case.special

        if cfg.case.r0_type == "relaxed":
            assert (cfg.case.state0_path is not None) and ("r" in cfg.case.state0_keys)

        # get the config file name, e.g. "cases/db.yaml" -> "db"
        cfg.case.name = os.path.splitext(os.path.basename(cfg.config))[0]

    def initialize(self):
        """Initialize and return everything needed for the numerical setup.

        This function sets up all physical and numerical quantities including:

            - reference values and case specific parameters
            - integration time step
            - relaxation or simulation mode
            - equation of state
            - boundary conditions
            - state dictionary
            - function including external force and boundary conditions

        Returns: A tuple containing the following elements:

           - cfg (DictConfig): configuration arguments
           - box_size (ndarray): size of the simulation box starting at 0.0
           - state (dict): dictionary containing all field values
           - g_ext_fn (Callable): external force function
           - bc_fn (Callable): boundary conditions function (e.g. velocity at walls)
           - nw_fn (Callable): jit-able wall normal funct. when moving walls, else None
           - eos (Callable): equation of state function
           - key (PRNGKey): random key for sampling
           - displacement_fn (Callable): displacement function for edge features
           - shift_fn (Callable): shift function for position updates
        """

        cfg = self.cfg
        dx = cfg.case.dx
        dim = cfg.case.dim
        rho_ref = cfg.case.rho_ref
        viscosity = cfg.case.viscosity
        u_ref = cfg.case.u_ref
        cfl = cfg.solver.cfl

        key_prng = jax.random.PRNGKey(cfg.seed)

        # Primal: reference density, dynamic viscosity, and velocity
        # Derived: reference speed of sound, pressure
        c_ref = cfg.case.c_ref_factor * u_ref
        p_ref = rho_ref * c_ref**2 / cfg.eos.gamma
        # for free surface simulation p_background in the EoS has to be 0.0
        p_bg = cfg.eos.p_bg_factor * p_ref

        # calculate volume and mass
        h = dx
        volume_ref = h**dim
        mass_ref = volume_ref * rho_ref

        # time integration step dt
        dt_convective = cfl * h / (c_ref + u_ref)
        dt_viscous = cfl * h**2 * rho_ref / (viscosity + EPS)
        dt_body_force = cfl * (h / (cfg.case.g_ext_magnitude + EPS)) ** 0.5
        dt = np.amin([dt_convective, dt_viscous, dt_body_force]).item()
        # TODO: consider adaptive time step sizes

        print("dt_convective :", dt_convective)
        print("dt_viscous    :", dt_viscous)
        print("dt_body_force :", dt_body_force)
        print("dt_max_allowed:", dt)

        if cfg.solver.dt is not None:
            if cfg.solver.dt > dt:
                warnings.warn("Explicit dt should comply with CFL.", UserWarning)
            dt = cfg.solver.dt

        print("dt_final      :", dt)

        if cfg.case.mode == "rlx":
            # run a relaxation of randomly initialized state for 500 steps
            sequence_length = 5000
            cfg.solver.t_end = dt * sequence_length
            # turn background pressure on for homogeneous particle distribution
        else:
            sequence_length = int(cfg.solver.t_end / dt)

        # Equation of state
        if cfg.solver.name == "RIE":
            eos = RIEMANNEoS(rho_ref, p_bg, u_ref)
        else:
            eos = TaitEoS(p_ref, rho_ref, p_bg, cfg.eos.gamma)

        # initialize box and positions of particles
        if dim == 2:
            box_size = self._box_size2D(cfg.solver.n_walls)
            r, tag = self._init_pos2D(box_size, dx, cfg.solver.n_walls)
        elif dim == 3:
            box_size = self._box_size3D(cfg.solver.n_walls)
            r, tag = self._init_pos3D(box_size, dx, cfg.solver.n_walls)
        displacement_fn, shift_fn = space.periodic(side=box_size)

        num_particles = len(r)
        print("Total number of particles = ", num_particles)

        # add noise to the fluid particles to break symmetry
        key, subkey = jax.random.split(key_prng)
        if cfg.case.r0_noise_factor != 0.0:
            noise_std = cfg.case.r0_noise_factor * dx
            noise = get_noise_masked(r.shape, tag == Tag.FLUID, subkey, std=noise_std)
            # PBC: move all particles to the box limits after noise addition
            r = shift_fn(r, noise)

        # initialize the velocity given the coordinates r with the noise
        if dim == 2:
            v = vmap(self._init_velocity2D)(r)
        elif dim == 3:
            v = vmap(self._init_velocity3D)(r)

        # initialize all other field values
        rho, mass, eta, temperature, kappa, Cp = self._set_field_properties(
            num_particles, mass_ref, cfg.case
        )
        # whether to compute wall normals
        is_nw = (
            cfg.solver.free_slip or cfg.solver.name == "RIE"
        ) and cfg.solver.is_bc_trick
        # calculate wall normals if necessary
        nw = self._compute_wall_normals("scipy")(r, tag) if is_nw else jnp.zeros_like(r)

        # initialize the state dictionary
        state = {
            "r": r,
            "tag": tag,
            "u": v,
            "v": v,
            "dudt": jnp.zeros_like(v),
            "dvdt": jnp.zeros_like(v),
            "drhodt": jnp.zeros_like(rho),
            "rho": rho,
            "p": eos.p_fn(rho),
            "mass": mass,
            "eta": eta,
            "dTdt": jnp.zeros_like(rho),
            "T": temperature,
            "kappa": kappa,
            "Cp": Cp,
            "nw": nw,
        }

        # overwrite the state dictionary with the provided one
        if cfg.case.state0_path is not None:
            _state = read_h5(cfg.case.state0_path)
            for k in state:
                if k not in cfg.case.state0_keys:
                    continue
                assert k in _state, ValueError(f"Key {k} not found in state0 file.")
                assert state[k].shape == _state[k].shape, ValueError(
                    f"Shape mismatch for key {k} in state0 file."
                )
                state[k] = _state[k]

        # the following arguments are needed for dataset generation
        cfg.case.c_ref, cfg.case.p_ref, cfg.case.p_bg = c_ref, p_ref, p_bg
        cfg.solver.dt, cfg.solver.sequence_length = dt, sequence_length
        cfg.case.num_particles_max = num_particles
        cfg.case.pbc = [True, True, True]  # TODO: matscipy needs 3D
        cfg.case.bounds = np.array([np.zeros_like(box_size), box_size]).T.tolist()

        state = self._boundary_conditions_fn(state)

        g_ext_fn = self._external_acceleration_fn
        bc_fn = self._boundary_conditions_fn

        # whether to recompute the wall normals at every integration step
        is_nw_recompute = (tag == Tag.MOVING_WALL).any() and is_nw
        if is_nw_recompute:
            assert cfg.nl.backend != "matscipy", NotImplementedError(
                "Wall normals not yet implemented for matscipy neighbor list when "
                "working with moving boundaries. \nIf you work with moving boundaries, "
                "don't use one of: `nl.backend=matscipy` or `solver.free_slip=True` or "
                "`solver.name=RIE`."
            )
        kwargs = {"disp_fn": displacement_fn, "box_size": box_size, "state0": state}
        nw_fn = self._compute_wall_normals("jax", **kwargs) if is_nw_recompute else None

        return (
            cfg,
            box_size,
            state,
            g_ext_fn,
            bc_fn,
            nw_fn,
            eos,
            key,
            displacement_fn,
            shift_fn,
        )

    @abstractmethod
    def _box_size2D(self, n_walls):
        pass

    @abstractmethod
    def _box_size3D(self, n_walls):
        pass

    def _init_pos2D(self, box_size, dx, n_walls):
        r = pos_init_cartesian_2d(box_size, dx)
        tag = jnp.full(len(r), Tag.FLUID, dtype=int)
        return r, tag

    def _init_pos3D(self, box_size, dx, n_walls):
        r = pos_init_cartesian_3d(box_size, dx)
        tag = jnp.full(len(r), Tag.FLUID, dtype=int)
        return r, tag

    @abstractmethod
    def _init_walls_2d(self):
        """Create all solid walls of a 2D case."""
        pass

    @abstractmethod
    def _init_walls_3d(self):
        """Create all solid walls of a 3D case."""
        pass

    @abstractmethod
    def _init_velocity2D(self, r):
        pass

    @abstractmethod
    def _init_velocity3D(self, r):
        pass

    @abstractmethod
    def _external_acceleration_fn(self, r):
        pass

    @abstractmethod
    def _boundary_conditions_fn(self, state):
        pass

    def _get_relaxed_r0(self, box_size, dx):
        assert hasattr(self, "_load_only_fluid"), AttributeError

        cfg = self.cfg
        name = "_".join([cfg.case.name, str(cfg.case.dim), str(dx), str(cfg.seed)])
        init_path = os.path.join("data_relaxed", name + ".h5")

        if not os.path.isfile(init_path):
            message = (
                f"python main.py config={cfg.config} mode=rlx seed={str(cfg.seed)} "
                f"case.dim={str(cfg.case.dim)} case.dx={str(cfg.case.dx)} "
                f"solver.name=SPH solver.tvf=1 solver.r0_noise_factor=0.25 "
                f"io.write_type=['h5'] io.data_path=data_relaxed/"
            )
            raise FileNotFoundError(f"First execute this: \n{message}")

        state = read_h5(init_path)
        if self._load_only_fluid:
            return state["r"][state["tag"] == Tag.FLUID]
        else:
            return state["r"], state["tag"]

    def _set_field_properties(self, num_particles, mass_ref, case):
        rho = jnp.ones(num_particles) * case.rho_ref
        mass = jnp.ones(num_particles) * mass_ref
        eta = jnp.ones(num_particles) * case.viscosity
        temperature = jnp.ones(num_particles) * case.T_ref
        kappa = jnp.ones(num_particles) * case.kappa_ref
        Cp = jnp.ones(num_particles) * case.Cp_ref
        return rho, mass, eta, temperature, kappa, Cp

    def _set_default_rlx(self):
        """Set default values for relaxation case setup.

        These would only change if the domain is not full.
        """

        self._box_size2D_rlx = self._box_size2D
        self._box_size3D_rlx = self._box_size3D
        self._init_pos2D_rlx = self._init_pos2D
        self._init_pos3D_rlx = self._init_pos3D

    def _compute_wall_normals(self, backend="scipy", **kwargs):
        if self.cfg.case.dim == 2:
            wall_part_fn = self._init_walls_2d
        elif self.cfg.case.dim == 3:
            wall_part_fn = self._init_walls_3d
        else:
            raise NotImplementedError("1D wall BCs not yet implemented")

        if backend == "scipy":
            # If one makes `tag` static (-> `self.tag`), this function can be jitted.
            # But it is significatly slower than `backend="jax"` due to `pure_callback`.
            def body(r, tag):
                return compute_nws_scipy(
                    r,
                    tag,
                    self.cfg.case.dx,
                    self.cfg.solver.n_walls,
                    self.offset_vec,
                    wall_part_fn,
                )
        elif backend == "jax":
            # This implementation is used in the integrator when having moving walls.
            body = compute_nws_jax_wrapper(
                state0=kwargs["state0"],
                dx=self.cfg.case.dx,
                n_walls=self.cfg.solver.n_walls,
                offset_vec=self.offset_vec,
                box_size=kwargs["box_size"],
                pbc=self.cfg.case.pbc,
                cfg_nl=self.cfg.nl,
                displacement_fn=kwargs["disp_fn"],
                wall_part_fn=wall_part_fn,
            )

        return body


def set_relaxation(Case, cfg):
    """Make a relaxation case from a SimulationSetup instance.

    Create a child class of a particular SimulationSetup instance and overwrite:

        - _init_pos{2|3}D
        - _box_size{2|3}D
        - _tag{2|3}D
        - _init_velocity{2|3}D
        - _external_acceleration_fn
        - _boundary_conditions_fn
    """

    class Rlx(Case):
        """Relax particles in a box"""

        def __init__(self, cfg):
            super().__init__(cfg)

            # custom variables related only to this Simulation
            self.case.g_ext_magnitude = 0.0

            # use the relaxation setup from the main case
            self._init_pos2D = self._init_pos2D_rlx
            self._init_pos3D = self._init_pos3D_rlx
            self._box_size2D = self._box_size2D_rlx
            self._box_size3D = self._box_size3D_rlx

        def _init_velocity2D(self, r):
            return jnp.zeros_like(r)

        def _init_velocity3D(self, r):
            return self._init_velocity2D(r)

        def _external_acceleration_fn(self, r):
            return jnp.zeros_like(r)

        def _boundary_conditions_fn(self, state):
            mask1 = jnp.isin(state["tag"], wall_tags)[:, None]

            state["u"] = jnp.where(mask1, 0.0, state["u"])
            state["v"] = jnp.where(mask1, 0.0, state["v"])

            state["dudt"] = jnp.where(mask1, 0.0, state["dudt"])
            state["dvdt"] = jnp.where(mask1, 0.0, state["dvdt"])

            return state

    return Rlx(cfg)


def load_case(case_root: str, case_py_file: str) -> SimulationSetup:
    """Load Case class from Python file.

    Args:
        case_root (str): Path to the case root directory, e.g. "cases/".
        case_py_file (str): Name of the Python case file, e.g. "db.py".
    """

    file_path = os.path.join(case_root, case_py_file)
    spec = importlib.util.spec_from_file_location("case_module", file_path)
    case_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(case_module)

    # the class name has to be capital version of case_py_file without .py extension
    # e.g. "db.py" -> "DB"
    class_name = case_py_file.split(".")[0].upper()
    Case = getattr(case_module, class_name)

    return Case



"""Visualization tools for SPH simulations."""

import os

import matplotlib.pyplot as plt
import numpy as np
from omegaconf import OmegaConf

from jax_sph.io_state import read_h5
from jax_sph.utils import get_ekin


def plt_scatter(r, x, ttl, vmin=None, vmax=None):
    plt.figure()
    plt.scatter(r[:, 0], r[:, 1], c=np.array(x), vmin=vmin, vmax=vmax)
    plt.colorbar()
    plt.title(ttl)
    plt.show()


def plt_ekin(src_dir):
    # metadata
    cfg = OmegaConf.load(os.path.join(src_dir, "config.yaml"))

    run_name = src_dir.split("/")[-1]
    fig_root = os.path.join(*src_dir.split("/")[:-1])
    # split one long trajectory into train, valid, and test
    files = os.listdir(src_dir)
    files = [f for f in files if (".h5" in f)]
    files = sorted(files, key=lambda x: int(x.split("_")[1][:-3]))
    if len(files) > 10000:
        files = files[::10]

    e_kin = np.zeros(len(files))
    u_max = np.zeros(len(files))
    u_min = np.zeros(len(files))
    for j, filename in enumerate(files):
        file_path_h5 = os.path.join(src_dir, filename)
        state = read_h5(file_path_h5, array_type="numpy")
        e_kin[j] = get_ekin(state, cfg.case.dx)
        u_max[j] = state["u"].max()
        u_min[j] = state["u"].min()

    if "3D_TGV" in src_dir:
        e_kin /= (2 * np.pi) ** 3
    dEdt = (e_kin[1:] - e_kin[:-1]) / (cfg.solver.dt * cfg.io.write_every)
    Nx = round((cfg.case.bounds[0][1] - cfg.case.bounds[0][0]) / cfg.case.dx)

    fig, ax = plt.subplots(3, 1, figsize=(8, 10))
    ax[0].plot(e_kin, label=f"SPH Nx={Nx}")
    ax[0].set_title(r"$E_{kin}$")
    ax[1].plot(u_max, label=r"$u_{max}$")
    ax[1].plot(-u_min, label=r"$-u_{min}$")
    ax[1].legend()
    ax[1].set_title(r"$u_{max}$")
    ax[2].plot(dEdt, label=f"SPH Nx={Nx}")
    ax[2].set_title(r"$d(E_{kin})/dt$")
    if "tgv" in src_dir.lower():
        try:
            ax[0].set_yscale("log")

            Re = 1 / cfg.case.viscosity
            dEdt_ref = np.loadtxt(
                f"./validation/ref/tgv3d_ref_{int(Re)}.txt", delimiter=","
            )
            every_n = max(len(dEdt_ref) // 50, 1)
            x_ref = (
                dEdt_ref[::every_n, 0]
                / dEdt_ref[::every_n, 0].max()
                * len(dEdt)
                / (cfg.solver.t_end / 10)
            )
            ax[2].scatter(
                x_ref,
                dEdt_ref[::every_n, 1],
                s=20,
                edgecolors="k",
                lw=1,
                facecolors="none",
                label="Jax-fluids Nx=64",
            )
            ax[0].scatter(
                x_ref,
                dEdt_ref[::every_n, 2],
                s=20,
                edgecolors="k",
                lw=1,
                facecolors="none",
                label="Jax-fluids Nx=64",
            )
        except FileNotFoundError:
            print("Not working with TGV or no reference data available")

    ax[0].legend()
    ax[2].legend()

    for a in ax:
        a.grid()
    plt.tight_layout()

    plt.savefig(os.path.join(fig_root, f"{run_name}_plots.png"))
    plt.show()
    plt.close()

    print(f"Finished {src_dir}!")


if __name__ == "__main__":
    plt_ekin("notebooks/data/debug/2D_HT_SPH_123_20240525-020025")



"""Equation of state."""


from abc import ABC, abstractmethod


class BaseEoS(ABC):
    """Base class for SPH equation of state."""

    @abstractmethod
    def p_fn(self, rho: float):
        """Compute pressure from density."""
        pass

    @abstractmethod
    def rho_fn(self, p: float):
        """Compute density from pressure."""
        pass


class TaitEoS(BaseEoS):
    """Tait equation of state.

    From: "A generalized wall boundary condition for smoothed particle
    hydrodynamics", Adami et al 2012
    """

    def __init__(self, p_ref, rho_ref, p_background, gamma):
        self.p_ref = p_ref
        self.rho_ref = rho_ref
        self.p_bg = p_background
        self.gamma = gamma

    def p_fn(self, rho):
        return self.p_ref * ((rho / self.rho_ref) ** self.gamma - 1) + self.p_bg

    def rho_fn(self, p):
        p_temp = p + self.p_ref - self.p_bg
        return self.rho_ref * (p_temp / self.p_ref) ** (1 / self.gamma)


class RIEMANNEoS(BaseEoS):
    """Riemann SPH equation of state.

    From: "A weakly compressible SPH method based on a
    low-dissipation Riemann solver", Zhang, Hu, Adams, 2017
    """

    def __init__(self, rho_ref, p_background, u_ref):
        self.rho_ref = rho_ref
        self.u_ref = u_ref
        self.p_bg = p_background

    def p_fn(self, rho):
        return 100 * self.u_ref**2 * (rho - self.rho_ref) + self.p_bg

    def rho_fn(self, p):
        return (p - self.p_bg) / (100 * self.u_ref**2) + self.rho_ref



# Source: https://github.com/jax-md/jax-md
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code to transform functions on individual tuples of particles to sets."""

from enum import Enum, IntEnum
from functools import partial, reduce
from operator import mul
from typing import Any, Callable, Dict, Generator, Optional, Tuple, Union

import jax.numpy as jnp
import jraph
import numpy as onp
from absl import logging
from jax import eval_shape, jit, lax, ops, tree_map, vmap
from jax.core import ShapedArray

from jax_sph.jax_md import dataclasses, space, util

# Types


Array = util.Array
PyTree = Any
f32 = util.f32
f64 = util.f64

i32 = util.i32
i64 = util.i64

Box = space.Box
DisplacementOrMetricFn = space.DisplacementOrMetricFn
MetricFn = space.MetricFn
MaskFn = Callable[[Array], Array]


# Cell List


@dataclasses.dataclass
class CellList:
    """Stores the spatial partition of a system into a cell list.

    See :meth:`cell_list` for details on the construction / specification.
    Cell list buffers all have a common shape, S, where
      * `S = [cell_count_x, cell_count_y, cell_capacity]`
      * `S = [cell_count_x, cell_count_y, cell_count_z, cell_capacity]`
    in two- and three-dimensions respectively. It is assumed that each cell has
    the same capacity.

    Attributes:
      position_buffer: An ndarray of floating point positions with shape
        `S + [spatial_dimension]`.
      id_buffer: An ndarray of int32 particle ids of shape `S`. Note that empty
        slots are specified by `id = N` where `N` is the number of particles in
        the system.
      named_buffer: A dictionary of ndarrays of shape `S + [...]`. This contains
        side data placed into the cell list.
      did_buffer_overflow: A boolean specifying whether or not the cell list
        exceeded the maximum allocated capacity.
      cell_capacity: An integer specifying the maximum capacity of each cell in
        the cell list.
      update_fn: A function that updates the cell list at a fixed capacity.
    """

    position_buffer: Array
    id_buffer: Array
    named_buffer: Dict[str, Array]

    did_buffer_overflow: Array

    cell_capacity: int = dataclasses.static_field()
    cell_size: float = dataclasses.static_field()

    update_fn: Callable[..., "CellList"] = dataclasses.static_field()

    def update(self, position: Array, **kwargs) -> "CellList":
        cl_data = (self.cell_capacity, self.did_buffer_overflow, self.update_fn)
        return self.update_fn(position, cl_data, **kwargs)

    @property
    def kwarg_buffers(self):
        logging.warning(
            "kwarg_buffers renamed to named_buffer. The name "
            "kwarg_buffers will be depricated."
        )
        return self.named_buffer


@dataclasses.dataclass
class CellListFns:
    allocate: Callable[..., CellList] = dataclasses.static_field()
    update: Callable[
        [Array, Union[CellList, int]], CellList
    ] = dataclasses.static_field()

    def __iter__(self):
        return iter((self.allocate, self.update))


def _cell_dimensions(
    spatial_dimension: int, box_size: Box, minimum_cell_size: float
) -> Tuple[Box, Array, Array, int]:
    """Compute the number of cells-per-side and total number of cells in a box."""
    if isinstance(box_size, (int, float)):
        box_size = float(box_size)

    # NOTE(schsam): Should we auto-cast based on box_size? I can't imagine a case
    # in which the box_size would not be accurately represented by an f32.
    if isinstance(box_size, onp.ndarray) and (
        box_size.dtype == i32 or box_size.dtype == i64
    ):
        box_size = float(box_size)

    cells_per_side = onp.floor(box_size / minimum_cell_size)
    cell_size = box_size / cells_per_side
    cells_per_side = onp.array(cells_per_side, dtype=i32)

    if isinstance(box_size, (onp.ndarray, jnp.ndarray)):
        if box_size.ndim == 1 or box_size.ndim == 2:
            assert box_size.size == spatial_dimension
            flat_cells_per_side = onp.reshape(cells_per_side, (-1,))
            for cells in flat_cells_per_side:
                if cells < 3:
                    msg = (
                        "Box must be at least 3x the size of the grid spacing in each "
                        "dimension."
                    )
                    raise ValueError(msg)
            cell_count = reduce(mul, flat_cells_per_side, 1)
        elif box_size.ndim == 0:
            cell_count = cells_per_side**spatial_dimension
        else:
            raise ValueError(
                (
                    "Box must be either: a scalar, a vector, or a matrix. "
                    f"Found {box_size}."
                )
            )
    else:
        cell_count = cells_per_side**spatial_dimension

    return box_size, cell_size, cells_per_side, int(cell_count)


def count_cell_filling(
    position: Array, box_size: Box, minimum_cell_size: float
) -> Array:
    """Counts the number of particles per-cell in a spatial partition."""
    dim = int(position.shape[1])
    box_size, cell_size, cells_per_side, cell_count = _cell_dimensions(
        dim, box_size, minimum_cell_size
    )

    hash_multipliers = _compute_hash_constants(dim, cells_per_side)

    particle_index = jnp.array(position / cell_size, dtype=i32)
    particle_hash = jnp.sum(particle_index * hash_multipliers, axis=1)

    filling = ops.segment_sum(jnp.ones_like(particle_hash), particle_hash, cell_count)
    return filling


def _compute_hash_constants(spatial_dimension: int, cells_per_side: Array) -> Array:
    if cells_per_side.size == 1:
        return jnp.array(
            [[cells_per_side**d for d in range(spatial_dimension)]], dtype=i32
        )
    elif cells_per_side.size == spatial_dimension:
        one = jnp.array([[1]], dtype=i32)
        cells_per_side = jnp.concatenate((one, cells_per_side[:, :-1]), axis=1)
        return jnp.array(jnp.cumprod(cells_per_side), dtype=i32)
    else:
        raise ValueError()


def _neighboring_cells(dimension: int) -> Generator[onp.ndarray, None, None]:
    for dindex in onp.ndindex(*([3] * dimension)):
        yield onp.array(dindex, dtype=i32) - 1


def _estimate_cell_capacity(
    position: Array, box_size: Box, cell_size: float, buffer_size_multiplier: float
) -> int:
    cell_capacity = onp.max(count_cell_filling(position, box_size, cell_size))
    return int(cell_capacity * buffer_size_multiplier)


def shift_array(arr: Array, dindex: Array) -> Array:
    if len(dindex) == 2:
        dx, dy = dindex
        dz = 0
    elif len(dindex) == 3:
        dx, dy, dz = dindex

    if dx < 0:
        arr = jnp.concatenate((arr[1:], arr[:1]))
    elif dx > 0:
        arr = jnp.concatenate((arr[-1:], arr[:-1]))

    if dy < 0:
        arr = jnp.concatenate((arr[:, 1:], arr[:, :1]), axis=1)
    elif dy > 0:
        arr = jnp.concatenate((arr[:, -1:], arr[:, :-1]), axis=1)

    if dz < 0:
        arr = jnp.concatenate((arr[:, :, 1:], arr[:, :, :1]), axis=2)
    elif dz > 0:
        arr = jnp.concatenate((arr[:, :, -1:], arr[:, :, :-1]), axis=2)

    return arr


def unflatten_cell_buffer(arr: Array, cells_per_side: Array, dim: int) -> Array:
    if (
        isinstance(cells_per_side, (int, float))
        or util.is_array(cells_per_side)
        and not cells_per_side.shape
    ):
        cells_per_side = (int(cells_per_side),) * dim
    elif util.is_array(cells_per_side) and len(cells_per_side.shape) == 1:
        cells_per_side = tuple([int(x) for x in cells_per_side[::-1]])
    elif util.is_array(cells_per_side) and len(cells_per_side.shape) == 2:
        cells_per_side = tuple([int(x) for x in cells_per_side[0][::-1]])
    else:
        raise ValueError()
    return jnp.reshape(arr, cells_per_side + (-1,) + arr.shape[1:])


def cell_list(
    box_size: Box, minimum_cell_size: float, buffer_size_multiplier: float = 1.25
) -> CellListFns:
    r"""Returns a function that partitions point data spatially.

    Given a set of points :math:`\{x_i \in R^d\}` with associated data
    :math:`\{k_i \in R^m\}` it is often useful to partition the points / data
    spatially. A simple partitioning that can be implemented efficiently within
    XLA is a dense partition into a uniform grid called a cell list.

    Since XLA requires that shapes be statically specified inside of a JIT block,
    the cell list code can operate in two modes: allocation and update.

    Allocation creates a new cell list that uses a set of input positions to
    estimate the capacity of the cell list. This capacity can be adjusted by
    setting the `buffer_size_multiplier` or setting the `extra_capacity`.
    Allocation cannot be JIT.

    Updating takes a previously allocated cell list and places a new set of
    particles in the cells. Updating cannot resize the cell list and is therefore
    compatible with JIT. However, if the configuration has changed substantially
    it is possible that the existing cell list won't be large enough to
    accommodate all of the particles. In this case the `did_buffer_overflow` bit
    will be set to True.

    Args:
      box_size: A float or an ndarray of shape `[spatial_dimension]` specifying
        the size of the system. Note, this code is written for the case where the
        boundaries are periodic. If this is not the case, then the current code
        will be slightly less efficient.
      minimum_cell_size: A float specifying the minimum side length of each cell.
        Cells are enlarged so that they exactly fill the box.
      buffer_size_multiplier: A floating point multiplier that multiplies the
        estimated cell capacity to allow for fluctuations in the maximum cell
        occupancy.
    Returns:
      A `CellListFns` object that contains two methods, one to allocate the cell
      list and one to update the cell list. The update function can be called
      with either a cell list from which the capacity can be inferred or with
      an explicit integer denoting the capacity. Note that an existing cell list
      can also be updated by calling `cell_list.update(position)`.
    """

    if util.is_array(box_size):
        box_size = onp.array(box_size)
        if len(box_size.shape) == 1:
            box_size = onp.reshape(box_size, (1, -1))

    if util.is_array(minimum_cell_size):
        minimum_cell_size = onp.array(minimum_cell_size)

    def cell_list_fn(
        position: Array,
        capacity_overflow_update: Optional[
            Tuple[int, bool, Callable[..., CellList]]
        ] = None,
        extra_capacity: int = 0,
        **kwargs,
    ) -> CellList:
        N = position.shape[0]
        dim = position.shape[1]

        if dim != 2 and dim != 3:
            # NOTE(schsam): Do we want to check this in compute_fn as well?
            raise ValueError(
                f"Cell list spatial dimension must be 2 or 3. Found {dim}."
            )

        _, cell_size, cells_per_side, cell_count = _cell_dimensions(
            dim, box_size, minimum_cell_size
        )

        if capacity_overflow_update is None:
            cell_capacity = _estimate_cell_capacity(
                position, box_size, cell_size, buffer_size_multiplier
            )
            cell_capacity += extra_capacity
            overflow = False
            update_fn = cell_list_fn
        else:
            cell_capacity, overflow, update_fn = capacity_overflow_update

        hash_multipliers = _compute_hash_constants(dim, cells_per_side)

        # Create cell list data.
        particle_id = lax.iota(i32, N)
        # NOTE(schsam): We use the convention that particles that are successfully,
        # copied have their true id whereas particles empty slots have id = N.
        # Then when we copy data back from the grid, copy it to an array of shape
        # [N + 1, output_dimension] and then truncate it to an array of shape
        # [N, output_dimension] which ignores the empty slots.
        cell_position = jnp.zeros(
            (cell_count * cell_capacity, dim), dtype=position.dtype
        )
        cell_id = N * jnp.ones((cell_count * cell_capacity, 1), dtype=i32)

        # It might be worth adding an occupied mask. However, that will involve
        # more compute since often we will do a mask for species that will include
        # an occupancy test. It seems easier to design around this empty_data_value
        # for now and revisit the issue if it comes up later.
        empty_kwarg_value = 10**5
        cell_kwargs = {}
        #  pytype: disable=attribute-error
        for k, v in kwargs.items():
            if not util.is_array(v):
                raise ValueError(
                    (
                        f'Data must be specified as an ndarray. Found "{k}" '
                        f"with type {type(v)}."
                    )
                )
            if v.shape[0] != position.shape[0]:
                raise ValueError(
                    (
                        "Data must be specified per-particle (an ndarray "
                        f'with shape ({N}, ...)). Found "{k}" with '
                        f"shape {v.shape}."
                    )
                )
            kwarg_shape = v.shape[1:] if v.ndim > 1 else (1,)
            cell_kwargs[k] = empty_kwarg_value * jnp.ones(
                (cell_count * cell_capacity,) + kwarg_shape, v.dtype
            )
        #  pytype: enable=attribute-error
        indices = jnp.array(position / cell_size, dtype=i32)
        hashes = jnp.sum(indices * hash_multipliers, axis=1)

        # Copy the particle data into the grid. Here we use a trick to allow us to
        # copy into all cells simultaneously using a single lax.scatter call. To do
        # this we first sort particles by their cell hash. We then assign each
        # particle to have a cell id = hash * cell_capacity + grid_id where
        # grid_id is a flat list that repeats 0, .., cell_capacity. So long as
        # there are fewer than cell_capacity particles per cell, each particle is
        # guaranteed to get a cell id that is unique.
        sort_map = jnp.argsort(hashes)
        sorted_position = position[sort_map]
        sorted_hash = hashes[sort_map]
        sorted_id = particle_id[sort_map]

        sorted_kwargs = {}
        for k, v in kwargs.items():
            sorted_kwargs[k] = v[sort_map]

        sorted_cell_id = jnp.mod(lax.iota(i32, N), cell_capacity)
        sorted_cell_id = sorted_hash * cell_capacity + sorted_cell_id

        cell_position = cell_position.at[sorted_cell_id].set(sorted_position)
        sorted_id = jnp.reshape(sorted_id, (N, 1))
        cell_id = cell_id.at[sorted_cell_id].set(sorted_id)
        cell_position = unflatten_cell_buffer(cell_position, cells_per_side, dim)
        cell_id = unflatten_cell_buffer(cell_id, cells_per_side, dim)

        for k, v in sorted_kwargs.items():
            if v.ndim == 1:
                v = jnp.reshape(v, v.shape + (1,))
            cell_kwargs[k] = cell_kwargs[k].at[sorted_cell_id].set(v)
            cell_kwargs[k] = unflatten_cell_buffer(cell_kwargs[k], cells_per_side, dim)

        occupancy = ops.segment_sum(jnp.ones_like(hashes), hashes, cell_count)
        max_occupancy = jnp.max(occupancy)
        overflow = overflow | (max_occupancy > cell_capacity)

        return CellList(
            cell_position,
            cell_id,
            cell_kwargs,
            overflow,
            cell_capacity,
            cell_size,
            update_fn,
        )  # pytype: disable=wrong-arg-count

    def allocate_fn(position: Array, extra_capacity: int = 0, **kwargs) -> CellList:
        return cell_list_fn(position, extra_capacity=extra_capacity, **kwargs)

    def update_fn(
        position: Array, cl_or_capacity: Union[CellList, int], **kwargs
    ) -> CellList:
        if isinstance(cl_or_capacity, int):
            capacity = int(cl_or_capacity)
            return cell_list_fn(position, (capacity, False, cell_list_fn), **kwargs)
        cl = cl_or_capacity
        cl_data = (cl.cell_capacity, cl.did_buffer_overflow, cl.update_fn)
        return cell_list_fn(position, cl_data, **kwargs)

    return CellListFns(allocate_fn, update_fn)  # pytype: disable=wrong-arg-count


# Neighbor Lists


class PartitionErrorCode(IntEnum):
    """An enum specifying different error codes.

    Attributes:
      NONE: Means that no error was encountered during simulation.
      NEIGHBOR_LIST_OVERFLOW: Indicates that the neighbor list was not large
        enough to contain all of the particles. This should indicate that it is
        necessary to allocate a new neighbor list.
      CELL_LIST_OVERFLOW: Indicates that the cell list was not large enough to
        contain all of the particles. This should indicate that it is necessary
        to allocate a new cell list.
      CELL_SIZE_TOO_SMALL: Indicates that the size of cells in a cell list was
        not large enough to properly capture particle interactions. This
        indicates that it is necessary to allcoate a new cell list with larger
        cells.
      MALFORMED_BOX: Indicates that a box matrix was not properly upper
        triangular.
    """

    NONE = 0
    NEIGHBOR_LIST_OVERFLOW = 1 << 0
    CELL_LIST_OVERFLOW = 1 << 1
    CELL_SIZE_TOO_SMALL = 1 << 2
    MALFORMED_BOX = 1 << 3


PEC = PartitionErrorCode


@dataclasses.dataclass
class PartitionError:
    """A struct containing error codes while building / updating neighbor lists.

    Attributes:
      code: An array storing the error code. See `PartitionErrorCode` for
        details.
    """

    code: Array

    def update(self, bit: bytes, pred: Array) -> Array:
        """Possibly adds an error based on a predicate."""
        zero = jnp.zeros((), jnp.uint8)
        bit = jnp.array(bit, dtype=jnp.uint8)
        return PartitionError(self.code | jnp.where(pred, bit, zero))

    def __str__(self) -> str:
        """Produces a string representation of the error code."""
        if not jnp.any(self.code):
            return ""

        if jnp.any(self.code & PEC.NEIGHBOR_LIST_OVERFLOW):
            return "Partition Error: Neighbor list buffer overflow."

        if jnp.any(self.code & PEC.CELL_LIST_OVERFLOW):
            return "Partition Error: Cell list buffer overflow"

        if jnp.any(self.code & PEC.CELL_SIZE_TOO_SMALL):
            return "Partition Error: Cell size too small"

        if jnp.any(self.code & PEC.MALFORMED_BOX):
            return (
                "Partition Error: Incorrect box format. Expecting upper " "triangular."
            )

        raise ValueError(f"Unexpected Error Code {self.code}.")

    __repr__ = __str__


def _displacement_or_metric_to_metric_sq(
    displacement_or_metric: DisplacementOrMetricFn,
) -> MetricFn:
    """Checks whether or not a displacement or metric was provided."""
    for dim in range(1, 4):
        try:
            R = ShapedArray((dim,), f32)
            dR_or_dr = eval_shape(displacement_or_metric, R, R, t=0)
            if len(dR_or_dr.shape) == 0:
                return (
                    lambda Ra, Rb, **kwargs: displacement_or_metric(Ra, Rb, **kwargs)
                    ** 2
                )
            else:
                return lambda Ra, Rb, **kwargs: space.square_distance(
                    displacement_or_metric(Ra, Rb, **kwargs)
                )
        except TypeError:
            continue
        except ValueError:
            continue
    raise ValueError(
        "Canonicalize displacement not implemented for spatial dimension larger"
        "than 4."
    )


def _cell_size(box, minimum_cell_size) -> Array:
    cells_per_side = jnp.floor(box / minimum_cell_size)
    return box / cells_per_side


def _fractional_cell_size(box, cutoff):
    if jnp.isscalar(box) or box.ndim == 0:
        return cutoff / box
    elif box.ndim == 1:
        return cutoff / jnp.min(box)
    elif box.ndim == 2:
        if box.shape[0] == 1:
            return 1 / jnp.floor(box[0, 0] / cutoff)
        elif box.shape[0] == 2:
            xx = box[0, 0]
            yy = box[1, 1]
            xy = box[0, 1] / yy

            nx = xx / jnp.sqrt(1 + xy**2)
            ny = yy

            nmin = jnp.floor(jnp.min(jnp.array([nx, ny])) / cutoff)
            nmin = jnp.where(nmin == 0, 1, nmin)
            return 1 / nmin
        elif box.shape[0] == 3:
            xx = box[0, 0]
            yy = box[1, 1]
            zz = box[2, 2]
            xy = box[0, 1] / yy
            xz = box[0, 2] / zz
            yz = box[1, 2] / zz

            nx = xx / jnp.sqrt(1 + xy**2 + (xy * yz - xz) ** 2)
            ny = yy / jnp.sqrt(1 + yz**2)
            nz = zz

            nmin = jnp.floor(jnp.min(jnp.array([nx, ny, nz])) / cutoff)
            nmin = jnp.where(nmin == 0, 1, nmin)
            return 1 / nmin
        else:
            raise ValueError(
                "Expected box to be either 1-, 2-, or 3-dimensional "
                f"found {box.shape[0]}"
            )
    else:
        raise ValueError(
            "Expected box to be either a scalar, a vector, or a "
            f"matrix. Found {type(box)}."
        )


class NeighborListFormat(Enum):
    """An enum listing the different neighbor list formats.

    Attributes:
      Dense: A dense neighbor list where the ids are a square matrix
        of shape `(N, max_neighbors_per_atom)`. Here the capacity of the neighbor
        list must scale with the highest connectivity neighbor.
      Sparse: A sparse neighbor list where the ids are a rectangular
        matrix of shape `(2, max_neighbors)` specifying the start / end particle
        of each neighbor pair.
      OrderedSparse: A sparse neighbor list whose format is the same as `Sparse`
        where only bonds with i < j are included.
    """

    Dense = 0
    Sparse = 1
    OrderedSparse = 2


def is_sparse(fmt: NeighborListFormat) -> bool:
    return fmt is NeighborListFormat.Sparse or fmt is NeighborListFormat.OrderedSparse


def is_format_valid(fmt: NeighborListFormat):
    if fmt not in list(NeighborListFormat):
        raise ValueError(
            (
                "Neighbor list format must be a member of NeighborListFormat"
                f" found {fmt}."
            )
        )


def is_box_valid(box: Array) -> bool:
    if jnp.isscalar(box) or box.ndim == 0 or box.ndim == 1:
        return True
    if box.ndim == 2:
        return jnp.triu(box) == box
    return False


@dataclasses.dataclass
class NeighborList:
    """A struct containing the state of a Neighbor List.

    Attributes:
      idx: For an N particle system this is an `[N, max_occupancy]` array of
        integers such that `idx[i, j]` is the j-th neighbor of particle i.
      reference_position: The positions of particles when the neighbor list was
        constructed. This is used to decide whether the neighbor list ought to be
        updated.
      error: An error code that is used to identify errors that occured during
        neighbor list construction. See `PartitionError` and `PartitionErrorCode`
        for details.
      cell_list_capacity: An optional integer specifying the capacity of the cell
        list used as an intermediate step in the creation of the neighbor list.
      max_occupancy: A static integer specifying the maximum size of the
        neighbor list. Changing this will invoke a recompilation.
      format: A NeighborListFormat enum specifying the format of the neighbor
        list.
      cell_size: A float specifying the current minimum size of the cells used
        in cell list construction.
      cell_list_fn: The function used to construct the cell list.
      update_fn: A static python function used to update the neighbor list.
    """

    idx: Array
    reference_position: Array
    error: PartitionError
    cell_list_capacity: Optional[int] = dataclasses.static_field()
    max_occupancy: int = dataclasses.static_field()

    format: NeighborListFormat = dataclasses.static_field()
    cell_size: Optional[float] = dataclasses.static_field()
    cell_list_fn: Callable[[Array, CellList], CellList] = dataclasses.static_field()
    update_fn: Callable[
        [Array, "NeighborList"], "NeighborList"
    ] = dataclasses.static_field()

    def update(self, position: Array, **kwargs) -> "NeighborList":
        return self.update_fn(position, self, **kwargs)

    @property
    def did_buffer_overflow(self) -> bool:
        return self.error.code & (PEC.NEIGHBOR_LIST_OVERFLOW | PEC.CELL_LIST_OVERFLOW)

    @property
    def cell_size_too_small(self) -> bool:
        return self.error.code & PEC.CELL_SIZE_TOO_SMALL

    @property
    def malformed_box(self) -> bool:
        return self.error.code & PEC.MALFORMED_BOX


@dataclasses.dataclass
class NeighborListFns:
    """A struct containing functions to allocate and update neighbor lists.

    Attributes:
      allocate: A function to allocate a new neighbor list. This function cannot
        be compiled, since it uses the values of positions to infer the shapes.
      update: A function to update a neighbor list given a new set of positions
        and a previously allocated neighbor list.
    """

    allocate: Callable[..., NeighborList] = dataclasses.static_field()
    update: Callable[[Array, NeighborList], NeighborList] = dataclasses.static_field()

    def __call__(
        self,
        position: Array,
        neighbors: Optional[NeighborList] = None,
        extra_capacity: int = 0,
        **kwargs,
    ) -> NeighborList:
        """A function for backward compatibility with previous neighbor lists.

        Args:
          position: An `(N, dim)` array of particle positions.
          neighbors: An optional neighbor list object. If it is provided then
            the function updates the neighbor list, otherwise it allocates a new
            neighbor list.
          extra_capacity: Extra capacity to add if allocating the neighbor list.
        Returns:
          A neighbor list object.
        """
        logging.warning(
            "Using a deprecated code path to create / update neighbor "
            "lists. It will be removed in a later version of JAX MD. "
            "Using `neighbor_fn.allocate` and `neighbor_fn.update` "
            "is preferred."
        )
        if neighbors is None:
            return self.allocate(position, extra_capacity, **kwargs)
        return self.update(position, neighbors, **kwargs)

    def __iter__(self):
        return iter((self.allocate, self.update))


NeighborFn = Callable[[Array, Optional[NeighborList], Optional[int]], NeighborList]


def neighbor_list(
    displacement_or_metric: DisplacementOrMetricFn,
    box: Box,
    r_cutoff: float,
    dr_threshold: float = 0.0,
    capacity_multiplier: float = 1.25,
    disable_cell_list: bool = False,
    mask_self: bool = True,
    custom_mask_function: Optional[MaskFn] = None,
    fractional_coordinates: bool = False,
    format: NeighborListFormat = NeighborListFormat.Dense,
    **static_kwargs,
) -> NeighborFn:
    """Returns a function that builds a list neighbors for collections of points.

    Neighbor lists must balance the need to be jit compatible with the fact that
    under a jit the maximum number of neighbors cannot change (owing to static
    shape requirements). To deal with this, our `neighbor_list` returns a
    `NeighborListFns` object that contains two functions: 1)
    `neighbor_fn.allocate` create a new neighbor list and 2) `neighbor_fn.update`
    updates an existing neighbor list. Neighbor lists themselves additionally
    have a convenience `update` member function.

    Note that allocation of a new neighbor list cannot be jit compiled since it
    uses the positions to infer the maximum number of neighbors (along with
    additional space specified by the `capacity_multiplier`). Updating the
    neighbor list can be jit compiled; if the neighbor list capacity is not
    sufficient to store all the neighbors, the `did_buffer_overflow` bit
    will be set to `True` and a new neighbor list will need to be reallocated.

    Here is a typical example of a simulation loop with neighbor lists:

    .. code-block:: python

       init_fn, apply_fn = simulate.nve(energy_fn, shift, 1e-3)
       exact_init_fn, exact_apply_fn = simulate.nve(exact_energy_fn, shift, 1e-3)

       nbrs = neighbor_fn.allocate(R)
       state = init_fn(random.PRNGKey(0), R, neighbor_idx=nbrs.idx)

       def body_fn(i, state):
         state, nbrs = state
         nbrs = nbrs.update(state.position)
         state = apply_fn(state, neighbor_idx=nbrs.idx)
         return state, nbrs

       step = 0
       for _ in range(20):
         new_state, nbrs = lax.fori_loop(0, 100, body_fn, (state, nbrs))
         if nbrs.did_buffer_overflow:
           nbrs = neighbor_fn.allocate(state.position)
         else:
           state = new_state
           step += 1

    Args:
      displacement: A function `d(R_a, R_b)` that computes the displacement
        between pairs of points.
      box: Either a float specifying the size of the box, an array of
        shape `[spatial_dim]` specifying the box size for a cubic box in each
        spatial dimension, or a matrix of shape `[spatial_dim, spatial_dim]` that
        is _upper triangular_ and specifies the lattice vectors of the box.
      r_cutoff: A scalar specifying the neighborhood radius.
      dr_threshold: A scalar specifying the maximum distance particles can move
        before rebuilding the neighbor list.
      capacity_multiplier: A floating point scalar specifying the fractional
        increase in maximum neighborhood occupancy we allocate compared with the
        maximum in the example positions.
      disable_cell_list: An optional boolean. If set to `True` then the neighbor
        list is constructed using only distances. This can be useful for
        debugging but should generally be left as `False`.
      mask_self: An optional boolean. Determines whether points can consider
        themselves to be their own neighbors.
      custom_mask_function: An optional function. Takes the neighbor array
        and masks selected elements. Note: The input array to the function is
        `(n_particles, m)` where the index of particle 1 is in index in the first
        dimension of the array, the index of particle 2 is given by the value in
        the array
      fractional_coordinates: An optional boolean. Specifies whether positions
        will be supplied in fractional coordinates in the unit cube, :math:`[0, 1]^d`.
        If this is set to True then the `box_size` will be set to `1.0` and the
        cell size used in the cell list will be set to `cutoff / box_size`.
      format: The format of the neighbor list; see the :meth:`NeighborListFormat` enum
        for details about the different choices for formats. Defaults to `Dense`.
      **static_kwargs: kwargs that get threaded through the calculation of
        example positions.
    Returns:
      A NeighborListFns object that contains a method to allocate a new neighbor
      list and a method to update an existing neighbor list.
    """
    is_format_valid(format)
    box = lax.stop_gradient(box)
    r_cutoff = lax.stop_gradient(r_cutoff)
    dr_threshold = lax.stop_gradient(dr_threshold)

    box = f32(box)

    cutoff = r_cutoff + dr_threshold
    cutoff_sq = cutoff**2
    threshold_sq = (dr_threshold / f32(2)) ** 2
    metric_sq = _displacement_or_metric_to_metric_sq(displacement_or_metric)

    @partial(jit, static_argnums=0)
    def candidate_fn(positionShape) -> Array:
        candidates = jnp.arange(positionShape[0])
        return jnp.broadcast_to(
            candidates[None, :], (positionShape[0], positionShape[0])
        )

    @partial(jit, static_argnums=1)
    def cell_list_candidate_fn(cl_id_buffer, positionShape) -> Array:
        N, dim = positionShape

        idx = cl_id_buffer

        cell_idx = [idx]

        for dindex in _neighboring_cells(dim):
            if onp.all(dindex == 0):
                continue
            cell_idx += [shift_array(idx, dindex)]

        cell_idx = jnp.concatenate(cell_idx, axis=-2)
        cell_idx = cell_idx[..., jnp.newaxis, :, :]
        cell_idx = jnp.broadcast_to(cell_idx, idx.shape[:-1] + cell_idx.shape[-2:])

        def copy_values_from_cell(value, cell_value, cell_id):
            scatter_indices = jnp.reshape(cell_id, (-1,))
            cell_value = jnp.reshape(cell_value, (-1,) + cell_value.shape[-2:])
            return value.at[scatter_indices].set(cell_value)

        neighbor_idx = jnp.zeros((N + 1,) + cell_idx.shape[-2:], i32)
        neighbor_idx = copy_values_from_cell(neighbor_idx, cell_idx, idx)
        return neighbor_idx[:-1, :, 0]

    @jit
    def mask_self_fn(idx: Array) -> Array:
        self_mask = idx == jnp.reshape(
            jnp.arange(idx.shape[0], dtype=i32), (idx.shape[0], 1)
        )
        return jnp.where(self_mask, idx.shape[0], idx)

    @jit
    def prune_neighbor_list_dense(position: Array, idx: Array, **kwargs) -> Array:
        d = partial(metric_sq, **kwargs)
        d = space.map_neighbor(d)

        N = position.shape[0]
        neigh_position = position[idx]
        dR = d(position, neigh_position)

        mask = (dR < cutoff_sq) & (idx < N)
        out_idx = N * jnp.ones(idx.shape, i32)

        cumsum = jnp.cumsum(mask, axis=1)
        index = jnp.where(mask, cumsum - 1, idx.shape[1] - 1)
        p_index = jnp.arange(idx.shape[0])[:, None]
        out_idx = out_idx.at[p_index, index].set(idx)
        max_occupancy = jnp.max(cumsum[:, -1])

        return out_idx, max_occupancy

    @jit
    def prune_neighbor_list_sparse(position: Array, idx: Array, **kwargs) -> Array:
        d = partial(metric_sq, **kwargs)
        d = space.map_bond(d)

        N = position.shape[0]
        sender_idx = jnp.broadcast_to(jnp.arange(N)[:, None], idx.shape)

        sender_idx = jnp.reshape(sender_idx, (-1,))
        receiver_idx = jnp.reshape(idx, (-1,))
        dR = d(position[sender_idx], position[receiver_idx])

        mask = (dR < cutoff_sq) & (receiver_idx < N)
        if format is NeighborListFormat.OrderedSparse:
            mask = mask & (receiver_idx < sender_idx)

        out_idx = N * jnp.ones(receiver_idx.shape, i32)

        cumsum = jnp.cumsum(mask)
        index = jnp.where(mask, cumsum - 1, len(receiver_idx) - 1)
        receiver_idx = out_idx.at[index].set(receiver_idx)
        sender_idx = out_idx.at[index].set(sender_idx)
        max_occupancy = cumsum[-1]

        return jnp.stack((receiver_idx, sender_idx)), max_occupancy

    def neighbor_list_fn(
        position: Array, neighbors=None, extra_capacity: int = 0, **kwargs
    ) -> NeighborList:
        def neighbor_fn(position_and_error, max_occupancy=None):
            position, err = position_and_error
            N = position.shape[0]

            cl_fn = None
            cl = None
            cell_size = None
            if not disable_cell_list:
                if neighbors is None:
                    _box = kwargs.get("box", box)
                    cell_size = cutoff
                    if fractional_coordinates:
                        err = err.update(PEC.MALFORMED_BOX, is_box_valid(_box))
                        cell_size = _fractional_cell_size(_box, cutoff)
                        _box = 1.0
                    if jnp.all(cell_size < _box / 3.0):
                        cl_fn = cell_list(_box, cell_size, capacity_multiplier)
                        cl = cl_fn.allocate(position, extra_capacity=extra_capacity)
                else:
                    cell_size = neighbors.cell_size
                    cl_fn = neighbors.cell_list_fn
                    if cl_fn is not None:
                        cl = cl_fn.update(position, neighbors.cell_list_capacity)

            if cl is None:
                cl_capacity = None
                idx = candidate_fn(position.shape)
            else:
                err = err.update(PEC.CELL_LIST_OVERFLOW, cl.did_buffer_overflow)
                idx = cell_list_candidate_fn(cl.id_buffer, position.shape)
                cl_capacity = cl.cell_capacity

            if mask_self:
                idx = mask_self_fn(idx)
            if custom_mask_function is not None:
                idx = custom_mask_function(idx)

            if is_sparse(format):
                idx, occupancy = prune_neighbor_list_sparse(position, idx, **kwargs)
            else:
                idx, occupancy = prune_neighbor_list_dense(position, idx, **kwargs)

            if max_occupancy is None:
                _extra_capacity = (
                    extra_capacity if not is_sparse(format) else N * extra_capacity
                )
                max_occupancy = int(occupancy * capacity_multiplier + _extra_capacity)
                if max_occupancy > idx.shape[-1]:
                    max_occupancy = idx.shape[-1]
                if not is_sparse(format):
                    capacity_limit = N - 1 if mask_self else N
                elif format is NeighborListFormat.Sparse:
                    capacity_limit = N * (N - 1) if mask_self else N**2
                else:
                    capacity_limit = N * (N - 1) // 2
                if max_occupancy > capacity_limit:
                    max_occupancy = capacity_limit
            idx = idx[:, :max_occupancy]
            update_fn = neighbor_list_fn if neighbors is None else neighbors.update_fn
            return NeighborList(
                idx,
                position,
                err.update(PEC.NEIGHBOR_LIST_OVERFLOW, occupancy > max_occupancy),
                cl_capacity,
                max_occupancy,
                format,
                cell_size,
                cl_fn,
                update_fn,
            )  # pytype: disable=wrong-arg-count

        nbrs = neighbors
        if nbrs is None:
            return neighbor_fn((position, PartitionError(jnp.zeros((), jnp.uint8))))

        neighbor_fn = partial(neighbor_fn, max_occupancy=nbrs.max_occupancy)

        # If the box has been updated, then check that fractional coordinates are
        # enabled and that the cell list has big enough cells.
        if "box" in kwargs and not disable_cell_list:
            if not fractional_coordinates:
                raise ValueError(
                    "Neighbor list cannot accept a box keyword argument "
                    "if fractional_coordinates is not enabled."
                )
            # `cell_size` is really the minimum cell size.
            cur_cell_size = _cell_size(1.0, nbrs.cell_size)
            new_cell_size = _cell_size(
                1.0, _fractional_cell_size(kwargs["box"], cutoff)
            )
            err = nbrs.error.update(
                PEC.CELL_SIZE_TOO_SMALL, new_cell_size > cur_cell_size
            )
            err = err.update(PEC.MALFORMED_BOX, is_box_valid(kwargs["box"]))
            nbrs = dataclasses.replace(nbrs, error=err)

        d = partial(metric_sq, **kwargs)
        d = vmap(d)
        return lax.cond(
            jnp.any(d(position, nbrs.reference_position) > threshold_sq),
            (position, nbrs.error),
            neighbor_fn,
            nbrs,
            lambda x: x,
        )

    def allocate_fn(position: Array, extra_capacity: int = 0, **kwargs):
        return neighbor_list_fn(position, extra_capacity=extra_capacity, **kwargs)

    def update_fn(position: Array, neighbors, **kwargs):
        return neighbor_list_fn(position, neighbors, **kwargs)

    return NeighborListFns(allocate_fn, update_fn)  # pytype: disable=wrong-arg-count


def neighbor_list_mask(neighbor: NeighborList, mask_self: bool = False) -> Array:
    """Compute a mask for neighbor list."""
    if is_sparse(neighbor.format):
        mask = neighbor.idx[0] < len(neighbor.reference_position)
        if mask_self:
            mask = mask & (neighbor.idx[0] != neighbor.idx[1])
        return mask

    mask = neighbor.idx < len(neighbor.idx)
    if mask_self:
        N = len(neighbor.reference_position)
        self_mask = neighbor.idx != jnp.reshape(jnp.arange(N, dtype=i32), (N, 1))
        mask = mask & self_mask
    return mask


def to_jraph(
    neighbor: NeighborList,
    mask: Optional[Array] = None,
    nodes: Optional[PyTree] = None,
    edges: Optional[PyTree] = None,
    globals: Optional[PyTree] = None,
) -> jraph.GraphsTuple:
    """Convert a sparse neighbor list to a `jraph.GraphsTuple`.

    As in jraph, padding here is accomplished by adding a ficticious graph with a
    single node.

    Args:
      neighbor: A neighbor list that we will convert to the jraph format. Must be
        sparse.
      mask: An optional mask on the edges.

    Returns:
      A `jraph.GraphsTuple` that contains the topology of the neighbor list.
    """
    if not is_sparse(neighbor.format):
        raise ValueError(
            "Cannot convert a dense neighbor list to jraph format. "
            "Please use either NeighborListFormat.Sparse or "
            "NeighborListFormat.OrderedSparse."
        )

    receivers, senders = neighbor.idx
    N = len(neighbor.reference_position)

    _mask = neighbor_list_mask(neighbor)

    # Pad the nodes to add one fictitious node.
    def pad(x):
        padding = jnp.zeros((1,) + x.shape[1:], dtype=x.dtype)
        return jnp.concatenate((x, padding), axis=0)

    nodes = tree_map(pad, nodes)

    # Pad the globals to add one fictitious global.
    globals = tree_map(pad, globals)

    # If there is an additional mask, reorder the edges.
    if mask is not None:
        _mask = _mask & mask
        cumsum = jnp.cumsum(_mask)
        index = jnp.where(_mask, cumsum - 1, len(receivers))
        ordered = N * jnp.ones((len(receivers) + 1,), i32)
        receivers = ordered.at[index].set(receivers)[:-1]
        senders = ordered.at[index].set(senders)[:-1]

        def reorder_edges(x):
            return jnp.zeros_like(x).at[index].set(x)

        edges = tree_map(reorder_edges, edges)
        mask = receivers < N

    return jraph.GraphsTuple(
        nodes=nodes,
        edges=edges,
        receivers=receivers,
        senders=senders,
        globals=globals,
        n_node=jnp.array([N, 1]),
        n_edge=jnp.array([jnp.sum(_mask), jnp.sum(~_mask)]),
    )


def to_dense(neighbor: NeighborList) -> Array:
    """Converts a sparse neighbor list to dense ids. Cannot be JIT."""
    if neighbor.format is not Sparse:
        raise ValueError("Can only convert sparse neighbor lists to dense ones.")

    receivers, senders = neighbor.idx
    mask = neighbor_list_mask(neighbor)

    receivers = receivers[mask]
    senders = senders[mask]

    N = len(neighbor.reference_position)
    count = ops.segment_sum(jnp.ones(len(receivers), i32), receivers, N)
    max_count = jnp.max(count)
    offset = jnp.tile(jnp.arange(max_count), N)[: len(senders)]
    hashes = senders * max_count + offset
    dense_idx = N * jnp.ones((N * max_count,), i32)
    dense_idx = dense_idx.at[hashes].set(receivers).reshape((N, max_count))
    return dense_idx


Dense = NeighborListFormat.Dense
Sparse = NeighborListFormat.Sparse
OrderedSparse = NeighborListFormat.OrderedSparse



"""Default jax-sph configs."""

from omegaconf import DictConfig, OmegaConf


def set_defaults(cfg: DictConfig = OmegaConf.create({})) -> DictConfig:
    """Set default lagrangebench configs."""

    ### global and hardware-related configs

    # .yaml case configuration file
    cfg.config = None
    # Seed for random number generator
    cfg.seed = 123
    # Whether to disable jitting compilation
    cfg.no_jit = False
    # Which GPU to use. -1 for CPU
    cfg.gpu = 0
    # Data type. One of "float32" or "float64"
    cfg.dtype = "float64"
    # XLA memory fraction to be preallocated. The JAX default is 0.75.
    # Should be specified before importing the library.
    cfg.xla_mem_fraction = 0.75

    ### psysical case configuration
    cfg.case = OmegaConf.create({})

    # Wchich Python case file to configure.
    cfg.case.source = None
    # Simulation mode. One of "sim" (run simulation) or "rlx" (run relaxation)
    cfg.case.mode = "sim"
    # Dimension of the simulation. One of 2 or 3
    cfg.case.dim = 3
    # Average distance between particles [0.001, 0.1]
    cfg.case.dx = 0.05
    # Initial state h5 path. Overrides `r0_type`. Can be useful to restart a simulation.
    cfg.case.state0_path = None
    # Which properties to adopt from state0_path. Include all to restart a simulation.
    cfg.case.state0_keys = ["r"]
    # Position initialization type. One of "cartesian" or "relaxed". Cartesian can have
    # `r0_noise_factor` and relaxed requires a state to be present in `data_relaxed`.
    cfg.case.r0_type = "cartesian"
    # How much Gaussian noise to add to r0. ( _ * dx)
    cfg.case.r0_noise_factor = 0.0
    # Magnitude of external force field
    cfg.case.g_ext_magnitude = 0.0
    # Reference dynamic viscosity. Inversely proportional to Re.
    cfg.case.viscosity = 0.01
    # Estimate max flow velocity to calculate artificial speed of sound.
    cfg.case.u_ref = 1.0
    # Reference speed of sound factor w.r.t. u_ref.
    cfg.case.c_ref_factor = 10.0
    # Reference density
    cfg.case.rho_ref = 1.0
    # Reference temperature
    cfg.case.T_ref = 1.0
    # Reference thermal conductivity
    cfg.case.kappa_ref = 0.0
    # Reference heat capacity at constant pressure
    cfg.case.Cp_ref = 0.0
    # case-specific variables
    cfg.case.special = OmegaConf.create({})

    ### solver
    cfg.solver = OmegaConf.create({})

    # Solver name. One of "SPH" (standard SPH) or "RIE" (Riemann SPH)
    cfg.solver.name = "SPH"
    # Transport velocity inclusion factor [0,...,1]
    cfg.solver.tvf = 0.0
    # CFL condition factor
    cfg.solver.cfl = 0.25
    # Density evolution vs density summation
    cfg.solver.density_evolution = False
    # Density renormalization when density evolution
    cfg.solver.density_renormalize = False
    # Integration time step. If None, it is calculated from the CFL condition.
    cfg.solver.dt = None
    # Physical time length of simulation
    cfg.solver.t_end = 0.2
    # Parameter alpha of artificial viscosity term
    cfg.solver.artificial_alpha = 0.0
    # Whether to turn on free-slip boundary condition
    cfg.solver.free_slip = False
    # Riemann dissipation limiter parameter, -1 = off
    cfg.solver.eta_limiter = 3
    # Thermal conductivity (non-dimensional)
    cfg.solver.kappa = 0
    # Number of wall boundary particle layers
    cfg.solver.n_walls = 3
    # Whether to apply the heat conduction term
    cfg.solver.heat_conduction = False
    # Whether to apply boundaty conditions
    cfg.solver.is_bc_trick = False  # new

    ### kernel
    cfg.kernel = OmegaConf.create({})

    # Kernel name, choose one of:
    # "CSK" (cubic spline kernel)
    # "QSK" (quintic spline kernel)
    # "WC2K" (Wendland C2 kernel)
    # "WC4K" (Wendland C4 kernel)
    # "WC6K" (Wendland C4 kernel)
    # "GK" (gaussian kernel)
    # "SGK" (super gaussian kernel)
    cfg.kernel.name = "QSK"
    # Smoothing length factor
    cfg.kernel.h_factor = 1.0  # new. Should default to 1.3 WC2K and 1.0 QSK

    ### equation of state
    cfg.eos = OmegaConf.create({})

    # EoS name. One of "Tait" or "RIEMANN"
    cfg.eos.name = "Tait"
    # power in the Tait equation of state
    cfg.eos.gamma = 1.0
    # background pressure factor w.r.t. p_ref
    cfg.eos.p_bg_factor = 0.0

    ### neighbor list
    cfg.nl = OmegaConf.create({})

    # Neighbor list backend. One of "jaxmd_vmap", "jaxmd_scan", "matscipy"
    cfg.nl.backend = "jaxmd_vmap"
    # Number of partitions for neighbor list. Applies to jaxmd_scan only.
    cfg.nl.num_partitions = 1

    ### output writing
    cfg.io = OmegaConf.create({})

    # In which format to write states. A subset of ["h5", "vtk"]
    cfg.io.write_type = []
    # Every `write_every` step will be saved
    cfg.io.write_every = 1
    # Where to write and read data
    cfg.io.data_path = "./"
    # What to print to stdout. As list of possible properties.
    cfg.io.print_props = ["Ekin", "u_max"]

    return cfg


defaults = set_defaults()


def check_cfg(cfg: DictConfig) -> None:
    """Check if the configs are valid."""

    assert cfg.config is not None, "A configuration file must be specified."

    assert cfg.case.source is not None, "A Python case file must be specified."
    assert cfg.case.mode in ["sim", "rlx"], "Mode must be one of 'sim' or 'rlx'."
    assert cfg.case.dim in [2, 3], "Dimension must be 2 or 3."
    assert cfg.case.dx >= 0.0, "dx must be > 0.0."
    assert cfg.case.viscosity >= 0.0, "viscosity must be >= 0.0."
    _r0_types = ["cartesian", "relaxed"]
    assert cfg.case.r0_type in _r0_types, f"r0_type must be in {_r0_types}."
    assert cfg.case.r0_noise_factor >= 0.0, "r0_noise_factor must be >= 0.0."
    assert cfg.case.g_ext_magnitude is not None, "g_ext_magnitude must be specified."
    assert cfg.case.u_ref is not None, "u_ref must be specified."

    assert cfg.solver.name in ["SPH", "RIE"], "Solver must be one of 'SPH' or 'RIE'."
    assert 0.0 <= cfg.solver.tvf <= 1.0, "tvf must be in [0.0, 1.0]."
    assert cfg.solver.dt is None or cfg.solver.dt > 0.0, "dt must be > 0.0."
    assert cfg.solver.t_end > 0.0, "t_end must be > 0.0."
    assert cfg.solver.artificial_alpha >= 0.0, "artificial_alpha must be >= 0.0."
    assert (
        cfg.solver.eta_limiter == -1 or cfg.solver.eta_limiter >= 0
    ), "eta_limiter must be -1 or >0."
    assert cfg.solver.kappa >= 0.0, "kappa must be >= 0.0."

    assert cfg.kernel.name in ["QSK", "WC2K"], "Kernel must be one of 'QSK' or 'WC2K'."
    assert cfg.kernel.h_factor > 0.0, "h_factor must be > 0.0."

    _eos_name = ["Tait", "RIEMANN"]
    assert cfg.eos.name in _eos_name, f"EoS must be one of {_eos_name}."
    assert cfg.eos.gamma > 0.0, "gamma must be > 0.0."
    assert cfg.eos.p_bg_factor >= 0.0, "p_bg_factor must be >= 0.0."

    _nl_backends = ["jaxmd_vmap", "jaxmd_scan", "matscipy"]
    assert cfg.nl.backend in _nl_backends, f"nl_backend must be in {_nl_backends}."

    _io_trite_type = ["h5", "vtk"]
    assert all(
        [w in _io_trite_type for w in cfg.io.write_type]
    ), f"write_type must be in {_io_trite_type}."
    assert cfg.io.write_every > 0, "write_every must be > 0."
